from concurrent.futures import ThreadPoolExecutor
import json
import os
import threading
import time
from datetime import datetime, timezone
from queue import Empty, Queue
from typing import Type
from uuid import UUID

import grpc
from google.protobuf.struct_pb2 import Struct
from google.protobuf import empty_pb2
from pydantic import BaseModel
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    retry_if_exception_type,
    before_sleep_log,
)

from hyrex import constants
from hyrex.configs import TaskConfig
from hyrex.dispatcher.dispatcher import Dispatcher
from hyrex.env_vars import EnvVars
from hyrex.hyrex_queue import HyrexQueue
from hyrex.proto import gateway_pb2_grpc, requests_pb2, task_pb2
from hyrex.schemas import (
    CronJob,
    CronJobRun,
    DequeuedTask,
    EnqueueTaskRequest,
    QueuePattern,
    TaskRun,
    TaskStatus,
    WorkflowRunRequest,
)

# Define the epoch zero timestamp for comparison
EPOCH_ZERO = datetime(1970, 1, 1, tzinfo=timezone.utc)


def pydantic_aware_default(obj):
    if isinstance(obj, BaseModel):
        # If the object is a Pydantic model, call model_dump()
        # to get its dictionary representation. json.dumps can handle dicts.
        return obj.model_dump()
    # If it's not a Pydantic model and json.dumps doesn't know it,
    # let the default TypeError happen.
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


class PerformanceDispatcher(Dispatcher):
    PERFORMANCE_SERVER_HOST = os.getenv(EnvVars.PERF_SERVER_HOST, "api.hyrex.io")
    PERFORMANCE_SERVER_PORT = os.getenv(EnvVars.PERF_SERVER_PORT, "443")

    # Status mapping between Python TaskStatus enum and proto TaskStatus enum
    _PY_TO_PROTO_STATUS = {
        TaskStatus.success: task_pb2.TaskStatus.SUCCESS,
        TaskStatus.failed: task_pb2.TaskStatus.FAILED,
        TaskStatus.running: task_pb2.TaskStatus.RUNNING,
        TaskStatus.queued: task_pb2.TaskStatus.QUEUED,
        TaskStatus.up_for_cancel: task_pb2.TaskStatus.UP_FOR_CANCEL,
        TaskStatus.canceled: task_pb2.TaskStatus.CANCELED,
        TaskStatus.lost: task_pb2.TaskStatus.LOST,
        TaskStatus.stopped: task_pb2.TaskStatus.STOPPED,
        TaskStatus.skipped: task_pb2.TaskStatus.SKIPPED,
        TaskStatus.await_deps: task_pb2.TaskStatus.AWAIT_DEPS,
        TaskStatus.await_start_time: task_pb2.TaskStatus.AWAIT_START_TIME,
    }

    # Reverse mapping for proto to Python conversion
    _PROTO_TO_PY_STATUS = {v: k for k, v in _PY_TO_PROTO_STATUS.items()}

    # Priority mapping between Python int and proto Priority enum
    _PY_TO_PROTO_PRIORITY = {
        0: task_pb2.Priority.P_UNSPECIFIED,
        1: task_pb2.Priority.P1,
        2: task_pb2.Priority.P2,
        3: task_pb2.Priority.P3,
        4: task_pb2.Priority.P4,
        5: task_pb2.Priority.P5,
        6: task_pb2.Priority.P6,
        7: task_pb2.Priority.P7,
        8: task_pb2.Priority.P8,
        9: task_pb2.Priority.P9,
        10: task_pb2.Priority.P10,
    }

    def __init__(self, api_key: str, conn_string: str):
        # def __init__(self, api_key: str, batch_size=100, flush_interval=0.1):
        super().__init__()

        self.api_key = api_key
        self.api_key_metadata = (("x-api-key", self.api_key),)

        server_address = (
            f"{self.PERFORMANCE_SERVER_HOST}:{self.PERFORMANCE_SERVER_PORT}"
        )

        if os.getenv(EnvVars.LOCAL_TESTING):
            self.logger.info("Testing locally.")
            self.channel = grpc.insecure_channel(server_address)
        else:
            channel_credentials = grpc.ssl_channel_credentials()
            self.channel = grpc.secure_channel(server_address, channel_credentials)
        self.gateway_stub = gateway_pb2_grpc.GatewayServiceStub(self.channel)

        # TODO: Consider setting max workers specifically here.
        self.enqueue_executor = ThreadPoolExecutor()
        self.running = True

        self.register_shutdown_handlers()

    def register_app(self, app_info: dict):
        app_info_struct = Struct()
        app_info_struct.update(app_info)

        request_proto = requests_pb2.RegisterAppRequest()
        request_proto.app_info = app_info_struct

        try:
            response = self.gateway_stub.RegisterApp(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def _convert_enqueue_request_to_proto(
        self, task: EnqueueTaskRequest
    ) -> requests_pb2.EnqueueRequest:
        """
        Convert an EnqueueTaskRequest to a proto EnqueueRequest message.
        """
        proto_task = requests_pb2.EnqueueRequest()
        proto_task.id = str(task.id)
        proto_task.durable_id = str(task.id)
        proto_task.root_id = str(task.root_id)
        if task.parent_id:
            proto_task.parent_id = str(task.parent_id)
        if task.workflow_run_id:
            proto_task.workflow_run_id = str(task.workflow_run_id)
        if task.workflow_dependencies:
            proto_task.workflow_dependencies.extend(
                [str(dep) for dep in task.workflow_dependencies]
            )
        proto_task.task_name = task.task_name
        proto_task.queue = task.queue
        proto_task.max_retries = task.max_retries
        proto_task.priority = task.priority
        if task.timeout_seconds is not None:
            proto_task.timeout_seconds = task.timeout_seconds
        if task.idempotency_key:
            proto_task.idempotency_key = task.idempotency_key

        try:
            # json.dumps will handle dicts, lists, strings, numbers etc. directly.
            # If it encounters a Pydantic model (either as task.args itself or nested),
            # it will call our pydantic_aware_default function.
            json_string = json.dumps(task.args, default=pydantic_aware_default)
            proto_task.args = json_string.encode("utf-8")
        except TypeError as e:
            self.logger.error(f"Task {task.id}: Failed to serialize args to JSON: {e}")
            raise

        return proto_task

    @retry(
        stop=stop_after_attempt(4),  # Try up to 4 times (initial + 3 retries)
        wait=wait_exponential(
            multiplier=1, min=1, max=30
        ),  # Exponential backoff: 1s, 2s, 4s, 8s... capped at 30s
        retry=retry_if_exception_type(grpc.RpcError),
    )
    def _send_grpc_enqueue_request(self, proto_task: requests_pb2.EnqueueRequest):
        """
        Send a single gRPC enqueue request synchronously with automatic retry.
        This method is designed to be called from within the ThreadPoolExecutor.
        Tenacity handles the retry logic with exponential backoff.
        """
        try:
            response = self.gateway_stub.Enqueue(
                proto_task, metadata=self.api_key_metadata
            )
            return response
        except grpc.RpcError as e:
            # Only retry on certain error codes
            retryable_codes = [
                grpc.StatusCode.UNAVAILABLE,
                grpc.StatusCode.DEADLINE_EXCEEDED,
                grpc.StatusCode.RESOURCE_EXHAUSTED,
                grpc.StatusCode.ABORTED,
                grpc.StatusCode.INTERNAL,
            ]

            if e.code() not in retryable_codes:
                self.logger.error(
                    f"Non-retryable gRPC error: {e.code()} - {e.details()}"
                )
                raise  # This won't be retried by tenacity

            self.logger.warning(
                f"Retryable gRPC error: {e.code()} - {e.details()}. Retrying..."
            )
            raise  # This will be retried by tenacity

    def _send_grpc_enqueue_callback(self, future):
        """
        Callback function to handle the result of an async gRPC request.
        """
        try:
            result = future.result()
            # Log success or process result as needed
            self.logger.debug(f"Enqueue request completed successfully")
        except Exception as e:
            # Log the error but don't re-raise to avoid crashing the executor
            self.logger.error(f"Enqueue request failed: {e}")

    def enqueue(self, tasks: list[EnqueueTaskRequest]):
        for task in tasks:
            proto_task = self._convert_enqueue_request_to_proto(task)
            # Submit the gRPC request to the thread pool for async execution
            future = self.enqueue_executor.submit(
                self._send_grpc_enqueue_request, proto_task
            )
            # Add callback to handle the result
            future.add_done_callback(self._send_grpc_enqueue_callback)

    def dequeue(
        self,
        executor_id: UUID,
        task_names: list[str],
        queue: str = constants.ANY_QUEUE,
        concurrency_limit: int = 0,
    ) -> DequeuedTask:
        request_proto = requests_pb2.DequeueRequest()
        request_proto.queue = queue
        request_proto.executor_id = str(executor_id)

        try:
            response = self.gateway_stub.Dequeue(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(f"gRPC call successful, response {response}")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

        # No task found
        # Check for task.id instead of just task because sometimes Python parses this as an instantiated empty task
        if not response.task_run.id:
            return None

        task_run = response.task_run

        # Parse args from bytes to dict
        args_dict = {}
        if task_run.args:
            try:
                args_dict = json.loads(task_run.args.decode("utf-8"))
            except (json.JSONDecodeError, UnicodeDecodeError) as e:
                self.logger.error(f"Failed to decode task args: {e}")

        # Convert protobuf timestamp to datetime
        queued_time = datetime.fromtimestamp(task_run.queued.seconds)
        started_time = datetime.fromtimestamp(task_run.started.seconds)

        scheduled_start = None
        if task_run.HasField("scheduled_start"):
            scheduled_start = datetime.fromtimestamp(task_run.scheduled_start.seconds)

        timeout_seconds = None
        if task_run.HasField("timeout_seconds"):
            timeout_seconds = task_run.timeout_seconds

        workflow_run_id = None
        if task_run.HasField("workflow_run_id"):
            workflow_run_id = UUID(task_run.workflow_run_id)

        parent_id = None
        if task_run.parent_id:
            parent_id = UUID(task_run.parent_id)

        return DequeuedTask(
            id=UUID(task_run.id),
            durable_id=UUID(task_run.durable_id),
            root_id=UUID(task_run.root_id),
            parent_id=parent_id,
            task_name=task_run.task_name,
            args=args_dict,
            queue=task_run.queue,
            priority=task_run.priority,
            timeout_seconds=timeout_seconds,
            scheduled_start=scheduled_start,
            queued=queued_time,
            started=started_time,
            workflow_run_id=workflow_run_id,
            attempt_number=task_run.attempt_number,
            max_retries=task_run.max_retries,
        )

    # def enqueue(
    #     self,
    #     tasks: list[EnqueueTaskRequest],
    # ):
    #     for task in tasks:
    #         self.local_queue.put(task)

    # def _batch_enqueue(self):
    #     tasks = []
    #     last_flush_time = time.monotonic()
    #     while self.running or not self.local_queue.empty():
    #         timeout = self.flush_interval - (time.monotonic() - last_flush_time)
    #         if timeout <= 0:
    #             # Flush if the flush interval has passed
    #             if tasks:
    #                 self._enqueue_tasks(tasks)
    #                 tasks = []
    #             last_flush_time = time.monotonic()
    #             continue

    #         try:
    #             # Wait for a task or until the timeout expires
    #             task = self.local_queue.get(timeout=timeout)
    #             tasks.append(task)
    #             if len(tasks) >= self.batch_size:
    #                 # Flush if batch size is reached
    #                 self._enqueue_tasks(tasks)
    #                 tasks = []
    #                 last_flush_time = time.monotonic()
    #         except Empty:
    #             # No task received within the timeout
    #             if tasks:
    #                 self._enqueue_tasks(tasks)
    #                 tasks = []
    #             last_flush_time = time.monotonic()

    #     # Flush any remaining tasks when stopping
    #     if tasks:
    #         self._enqueue_tasks(tasks)

    # def _enqueue_tasks(self, tasks: list[EnqueueTaskRequest]):
    #     enqueue_url = f"{self.HYREX_PLATFORM_URL}{self.ENQUEUE_TASK_PATH}"
    #     headers = {
    #         "X-API-Key": self.api_key,
    #     }

    #     task_list_json = [task.model_dump(mode="json") for task in tasks]

    #     try:
    #         response = requests.post(enqueue_url, headers=headers, json=task_list_json)
    #         if response.status_code != 200:
    #             self.logger.error(f"Error enqueuing task: {response.status_code}")
    #             self.logger.error(f"Response body: {response.text}")
    #     except requests.exceptions.RequestException as e:
    #         self.logger.error(f"Error enqueuing task via API: {str(e)}")
    #         raise RuntimeError(f"Failed to enqueue task via API: {e}")

    def stop(self):
        """
        Stops the batching process and flushes remaining tasks.
        """
        # Check if already stopping/stopped
        if not self.running:
            return

        self.logger.debug("Stopping dispatcher...")
        self.running = False
        self.enqueue_executor.shutdown(wait=True)
        if self.channel:
            self.channel.close()
        self.logger.debug("Dispatcher stopped successfully!")

    def mark_success(self, task_id: UUID, result: str):
        request_proto = requests_pb2.MarkSuccessRequest()
        request_proto.task_run_id = str(task_id)
        if result:
            request_proto.result = result

        try:
            self.gateway_stub.MarkSuccess(request_proto, metadata=self.api_key_metadata)
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC MarkSuccess call failed: {e.code()} - {e.details()}"
            )
            raise

    def mark_failed(self, task_id: UUID):
        request_proto = requests_pb2.MarkFailedRequest()
        request_proto.task_run_id = str(task_id)

        try:
            self.gateway_stub.MarkFailed(request_proto, metadata=self.api_key_metadata)
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC MarkFailed call failed: {e.code()} - {e.details()}"
            )
            raise

    def retry_task(self, task_id: UUID, backoff_seconds: int):
        request_proto = requests_pb2.RetryTaskRunRequest()
        request_proto.task_run_id = str(task_id)
        request_proto.backoff_seconds = backoff_seconds

        try:
            self.gateway_stub.RetryTaskRun(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC RetryTaskRun call failed: {e.code()} - {e.details()}"
            )
            raise

    # TODO: Implement
    def try_to_cancel_task(self, task_id: UUID):
        raise NotImplementedError("Cancellation not yet implemented on Hyrex platform")

    def task_canceled(self, task_id: UUID):
        request_proto = requests_pb2.MarkCanceledRequest()
        request_proto.task_run_id = str(task_id)

        try:
            self.gateway_stub.MarkCanceled(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("MarkCanceled gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC MarkCanceled call failed: {e.code()} - {e.details()}"
            )
            raise

    def get_task_status(self, task_id: UUID) -> TaskStatus:
        request_proto = requests_pb2.GetTaskRunStatusRequest()
        request_proto.task_run_id = str(task_id)

        try:
            response = self.gateway_stub.GetTaskStatus(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(
                f"gRPC GetTaskStatus call successful, response: {response}"
            )
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC GetTaskStatus call failed: {e.code()} - {e.details()}"
            )
            raise

        return self._PROTO_TO_PY_STATUS[response.status]

    def register_executor(
        self,
        executor_id: UUID,
        executor_name: str,
        queue_pattern: str,
        queues: list[HyrexQueue],
        worker_name: str,
    ):
        request_proto = requests_pb2.RegisterExecutorRequest()
        request_proto.executor_id = str(executor_id)
        request_proto.executor_name = executor_name
        request_proto.queue_pattern = queue_pattern
        request_proto.queues.extend([queue.name for queue in queues])
        request_proto.worker_name = worker_name

        try:
            self.gateway_stub.RegisterExecutor(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC RegisterExecutor call failed: {e.code()} - {e.details()}"
            )
            raise

    def disconnect_executor(self, executor_id: UUID):
        request_proto = requests_pb2.DisconnectExecutorRequest()
        request_proto.executor_id = str(executor_id)

        try:
            self.gateway_stub.DisconnectExecutor(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC DisconnectExecutor call failed: {e.code()} - {e.details()}"
            )
            raise

    def mark_running_tasks_lost(self, executor_id: UUID):
        request_proto = requests_pb2.MarkRunningTasksLostRequest()
        request_proto.executor_id = str(executor_id)

        try:
            self.gateway_stub.MarkRunningTasksLost(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("MarkRunningTasksLost gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC MarkRunningTasksLost call failed: {e.code()} - {e.details()}"
            )
            raise

    def executor_heartbeat(self, executor_ids: list[UUID], timestamp: datetime):
        request_proto = requests_pb2.ExecutorHeartbeatRequest()
        request_proto.executor_ids.extend(
            [str(executor_id) for executor_id in executor_ids]
        )

        # Convert datetime to protobuf timestamp
        request_proto.timestamp.FromDatetime(timestamp)

        try:
            self.gateway_stub.ExecutorHeartbeat(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("ExecutorHeartbeat gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC ExecutorHeartbeat call failed: {e.code()} - {e.details()}"
            )
            raise

    def update_executor_stats(self, executor_id: UUID, stats: dict):
        request_proto = requests_pb2.UpdateExecutorStatsRequest()
        request_proto.executor_id = str(executor_id)

        # Convert dict to protobuf Struct
        stats_struct = Struct()
        stats_struct.update(stats)
        request_proto.executor_stats.CopyFrom(stats_struct)

        try:
            self.gateway_stub.UpdateExecutorStats(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("UpdateExecutorStats gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC UpdateExecutorStats call failed: {e.code()} - {e.details()}"
            )
            raise

    def task_heartbeat(self, task_ids: list[UUID], timestamp: datetime):
        request_proto = requests_pb2.TaskRunHeartbeatRequest()
        request_proto.task_run_ids.extend([str(task_id) for task_id in task_ids])

        # Convert datetime to protobuf timestamp
        request_proto.timestamp.FromDatetime(timestamp)

        try:
            self.gateway_stub.TaskRunHeartbeat(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("TaskRunHeartbeat gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC TaskRunHeartbeat call failed: {e.code()} - {e.details()}"
            )
            raise

    def get_tasks_up_for_cancel(self) -> list[UUID]:
        try:
            response = self.gateway_stub.GetTaskRunsUpForCancel(
                empty_pb2.Empty(), metadata=self.api_key_metadata
            )
            self.logger.debug(
                f"GetTaskRunsUpForCancel gRPC call successful, response: {response}"
            )

            # Convert string UUIDs to UUID objects
            return [UUID(task_id) for task_id in response.task_run_ids]
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC GetTaskRunsUpForCancel call failed: {e.code()} - {e.details()}"
            )
            raise

    def get_queues_for_pattern(self, pattern: QueuePattern) -> list[str]:
        request_proto = requests_pb2.GetQueuesRequest()
        request_proto.max_num_queues = 10000
        request_proto.pattern = pattern.glob_pattern

        try:
            response = self.gateway_stub.GetQueues(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(f"GetQueues gRPC call successful. response: {response}")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

        return response.queues

    def register_task_def(
        self,
        task_name: str,
        arg_schema: Type[BaseModel] | None,
        task_config: TaskConfig,
        cron: str = None,
        source_code: str = None,
    ):
        # The proto structure has changed to use a Task message
        request_proto = requests_pb2.RegisterTaskDefRequest()

        # Create the TaskDef message
        task_def = task_pb2.TaskDef()
        task_def.task_name = task_name

        # Handle arg_schema
        if arg_schema and hasattr(arg_schema, "model_json_schema"):
            arg_schema_struct = Struct()
            arg_schema_struct.update(arg_schema.model_json_schema())
            task_def.arg_schema.CopyFrom(arg_schema_struct)

        # Set required fields from task_config
        task_def.queue = (
            task_config.get_queue_name() if task_config.queue else "default"
        )
        task_def.max_retries = (
            task_config.max_retries if task_config.max_retries is not None else 0
        )

        # Map Python priority to protobuf Priority enum
        priority_value = task_config.priority if task_config.priority is not None else 5
        task_def.priority = self._PY_TO_PROTO_PRIORITY.get(
            priority_value, task_pb2.Priority.P5
        )

        # Set optional fields
        if task_config.timeout_seconds is not None:
            task_def.timeout_seconds = task_config.timeout_seconds

        if cron:
            task_def.cron = cron

        if source_code:
            task_def.source_code = source_code

        # Set the task in the request proto
        request_proto.task_def.CopyFrom(task_def)

        try:
            self.gateway_stub.RegisterTaskDef(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("gRPC RegisterTaskDef call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC RegisterTaskDef call failed: {e.code()} - {e.details()}"
            )
            raise

    def acquire_scheduler_lock(self, worker_name: str) -> int | None:
        pass

    def pull_cron_job_expressions(self) -> list[CronJob]:
        return []

    def update_cron_job_confirmation_timestamp(self, jobid: int):
        pass

    def schedule_cron_job_runs(self, cron_job_runs: list[CronJobRun]):
        pass

    def register_cron_sql_query(
        self,
        cron_job_name: str,
        cron_sql_query: str,
        cron_expr: str,
        should_backfill: bool,
    ) -> None:
        pass

    def execute_queued_cron_job_run(self) -> str | None:
        pass

    def release_scheduler_lock(self, worker_name: str) -> None:
        pass

    def register_workflow(
        self,
        name: str,
        source_code: str,
        workflow_dag_json: str,
        workflow_arg_schema: Type[BaseModel] | None,
        default_config: dict,
        cron: str | None,
    ):
        request_proto = requests_pb2.RegisterWorkflowRequest()
        request_proto.workflow_name = name
        request_proto.source_code = source_code

        # Handle workflow_dag_json - convert dict to JSON string if needed
        if isinstance(workflow_dag_json, dict):
            request_proto.workflow_dag_json = json.dumps(workflow_dag_json)
        else:
            request_proto.workflow_dag_json = workflow_dag_json

        # Handle arg_schema
        if workflow_arg_schema:
            arg_schema_struct = Struct()
            # Convert Pydantic BaseModel class to JSON schema dict
            if hasattr(workflow_arg_schema, "model_json_schema"):
                schema_dict = workflow_arg_schema.model_json_schema()
                arg_schema_struct.update(schema_dict)
            else:
                # If it's already a dict or other type, use it directly
                arg_schema_struct.update(workflow_arg_schema)
            request_proto.workflow_arg_schema.CopyFrom(arg_schema_struct)

        # Handle default_config
        if default_config:
            default_config_struct = Struct()
            default_config_struct.update(default_config)
            request_proto.default_config.CopyFrom(default_config_struct)

        # Handle cron
        if cron:
            request_proto.cron = cron

        try:
            self.gateway_stub.RegisterWorkflow(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("gRPC RegisterWorkflow call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC RegisterWorkflow call failed: {e.code()} - {e.details()}"
            )
            raise

    def send_workflow_run(self, workflow_run_request: WorkflowRunRequest) -> UUID:
        request_proto = requests_pb2.SendWorkflowRunRequest()
        request_proto.workflow_run_id = str(workflow_run_request.id)
        request_proto.workflow_name = workflow_run_request.workflow_name
        request_proto.queue = workflow_run_request.queue

        # Convert args to protobuf Struct
        args_struct = Struct()
        args_struct.update(workflow_run_request.args)
        request_proto.args.CopyFrom(args_struct)

        # Handle optional fields
        if workflow_run_request.timeout_seconds is not None:
            request_proto.timeout_seconds = workflow_run_request.timeout_seconds

        if workflow_run_request.idempotency_key:
            request_proto.idempotency_key = workflow_run_request.idempotency_key

        try:
            self.gateway_stub.SendWorkflowRun(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("gRPC SendWorkflowRun call successful")
            return workflow_run_request.id
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC SendWorkflowRun call failed: {e.code()} - {e.details()}"
            )
            raise

    def advance_workflow_run(self, workflow_run_id: UUID):
        request_proto = requests_pb2.AdvanceWorkflowRunRequest()
        request_proto.workflow_run_id = str(workflow_run_id)

        try:
            self.gateway_stub.AdvanceWorkflowRun(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("gRPC AdvanceWorkflowRun call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC AdvanceWorkflowRun call failed: {e.code()} - {e.details()}"
            )
            raise

    def get_workflow_run_args(self, workflow_run_id: UUID) -> dict:
        request_proto = requests_pb2.GetWorkflowRunArgsRequest()
        request_proto.workflow_run_id = str(workflow_run_id)

        try:
            response = self.gateway_stub.GetWorkflowRunArgs(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("gRPC GetWorkflowRunArgs call successful")

            # Convert protobuf Struct to dict
            return dict(response.args)
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC GetWorkflowRunArgs call failed: {e.code()} - {e.details()}"
            )
            raise

    def get_durable_run_tasks(self, durable_id: UUID) -> list[TaskRun]:
        request_proto = requests_pb2.GetDurableTaskRunsRequest()
        request_proto.durable_id = str(durable_id)

        try:
            response = self.gateway_stub.GetDurableTaskRuns(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(f"gRPC call successful, response: {response}")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

        # This helps handle immediate `wait()` calls when task is sent, but not yet processed by performance server.
        # TODO: Handle this better!
        if not response.task_runs:
            time.sleep(0.5)
            return self.get_durable_run_tasks(durable_id=durable_id)

        python_task_runs = []
        for proto_task in response.task_runs:
            try:
                # Map protobuf enum to Python StrEnum using the mapping dict
                current_status = self._PROTO_TO_PY_STATUS.get(proto_task.status)
                if current_status is None:
                    # Fallback: try to convert protobuf enum name to Python enum
                    status_name = task_pb2.TaskStatus.Name(proto_task.status)
                    self.logger.warning(
                        f"Unknown proto status {proto_task.status} ({status_name})"
                    )
                    current_status = TaskStatus(
                        status_name
                    )  # This will raise if not found

                # Convert timestamps, checking for epoch zero (default/unset) for optional Python fields
                queued_dt = proto_task.queued.ToDatetime(
                    tzinfo=timezone.utc
                )  # Store with UTC timezone
                started_dt = proto_task.started.ToDatetime(tzinfo=timezone.utc)
                finished_dt = proto_task.finished.ToDatetime(tzinfo=timezone.utc)

                queued_val = queued_dt if queued_dt > EPOCH_ZERO else None
                started_val = started_dt if started_dt > EPOCH_ZERO else None
                finished_val = finished_dt if finished_dt > EPOCH_ZERO else None

                # Parse JSON result string
                task_result = (
                    json.loads(proto_task.result) if proto_task.result else None
                )

                python_task = TaskRun(
                    id=UUID(proto_task.id),
                    task_name=proto_task.task_name,
                    max_retries=proto_task.max_retries,
                    attempt_number=proto_task.attempt_number,
                    status=current_status,
                    queued=queued_val,
                    started=started_val,
                    finished=finished_val,
                    result=task_result,
                )
                python_task_runs.append(python_task)

            except (ValueError, json.JSONDecodeError, KeyError) as e:
                # Handle potential errors during conversion (e.g., invalid UUID, bad JSON, invalid status enum)
                self.logger.error(f"Error converting TaskRun {proto_task.id}: {e}")
                # TODO: Decide how to proceed: skip this task, raise an error, etc.
                raise

        return python_task_runs

    def get_workflow_durable_runs(self, workflow_run_id: UUID) -> list[UUID]:
        request_proto = requests_pb2.GetWorkflowDurableRunsRequest()
        request_proto.workflow_run_id = str(workflow_run_id)

        try:
            response = self.gateway_stub.GetWorkflowDurableRuns(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("gRPC GetWorkflowDurableRuns call successful")

            # Convert string UUIDs to UUID objects
            return [UUID(durable_id) for durable_id in response.durable_ids]
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC GetWorkflowDurableRuns call failed: {e.code()} - {e.details()}"
            )
            raise

    def try_to_cancel_durable_run(self, durable_id: UUID):
        request_proto = requests_pb2.TryToCancelDurableRunRequest()
        request_proto.durable_id = str(durable_id)

        try:
            self.gateway_stub.TryToCancelDurableRun(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("TryToCancelDurableRun gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC TryToCancelDurableRun call failed: {e.code()} - {e.details()}"
            )
            raise

    def update_executor_queues(self, executor_id: UUID, queues: list[str]):
        request_proto = requests_pb2.UpdateExecutorQueuesRequest()
        request_proto.executor_id = str(executor_id)
        request_proto.queues.extend(
            queues
        )  # Use extend for repeated field instead of direct assignment

        try:
            self.gateway_stub.UpdateExecutorQueues(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def save_result(self, task_id: UUID, result: str):
        request_proto = requests_pb2.SaveTaskRunResultRequest()
        request_proto.task_run_id = str(task_id)
        request_proto.result = result

        try:
            self.gateway_stub.SaveTaskResult(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def get_result(self, task_id: UUID) -> dict:
        request_proto = requests_pb2.GetTaskRunResultRequest()
        request_proto.task_run_id = str(task_id)

        try:
            response = self.gateway_stub.GetTaskResult(
                request_proto, metadata=self.api_key_metadata
            )
            return json.loads(response.result)
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def set_log_link(self, task_id: UUID, log_link: str):
        request_proto = requests_pb2.SetLogLinkRequest()
        request_proto.task_run_id = str(task_id)
        request_proto.log_link = log_link

        try:
            self.gateway_stub.SetLogLink(request_proto, metadata=self.api_key_metadata)
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def write_s3_logs(self, task_id: UUID, logs: str):
        request_proto = requests_pb2.WriteLogsRequest()
        request_proto.task_run_id = str(task_id)
        request_proto.logs = logs

        try:
            self.gateway_stub.WriteLogs(request_proto, metadata=self.api_key_metadata)
            self.logger.debug("WriteLogs gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC WriteLogs call failed: {e.code()} - {e.details()}")
            raise

    def kv_set(self, key: str, value: str) -> None:
        request_proto = requests_pb2.KVStoreSetRequest()
        request_proto.key = key
        request_proto.value = value
        request_proto.overwrite = True  # Always overwrite for now

        try:
            self.gateway_stub.KVStoreSet(request_proto, metadata=self.api_key_metadata)
            self.logger.debug("KVStoreSet gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC KVStoreSet call failed: {e.code()} - {e.details()}")
            raise

    def kv_get(self, key: str) -> str | None:
        request_proto = requests_pb2.KVStoreGetRequest()
        request_proto.key = key

        try:
            response = self.gateway_stub.KVStoreGet(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("KVStoreGet gRPC call successful")
            return response.value if response.value else None
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.NOT_FOUND:
                return None
            self.logger.error(f"gRPC KVStoreGet call failed: {e.code()} - {e.details()}")
            raise

    def kv_delete(self, key: str) -> None:
        request_proto = requests_pb2.KVStoreDeleteRequest()
        request_proto.key = key

        try:
            self.gateway_stub.KVStoreDelete(request_proto, metadata=self.api_key_metadata)
            self.logger.debug("KVStoreDelete gRPC call successful")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC KVStoreDelete call failed: {e.code()} - {e.details()}")
            raise



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: acquire_scheduler_lock.sql
import dataclasses
import datetime
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


ACQUIRE_SCHEDULER_LOCK = """-- name: acquire_scheduler_lock \\:one
INSERT INTO hyrex_scheduler_lock (
    lockid, worker_name, acquired_at, heartbeat_at, release_at, is_active
)
VALUES (
    1,
    :p1,
    now(),
    now(),
    now() + CAST(:p2 AS interval),
    true
)
ON CONFLICT (lockid)
  DO UPDATE 
     SET worker_name  = EXCLUDED.worker_name,
         acquired_at  = EXCLUDED.acquired_at,
         heartbeat_at = EXCLUDED.heartbeat_at,
         release_at   = EXCLUDED.release_at,
         is_active    = EXCLUDED.is_active
   WHERE (
       -- Only update if the lock is not truly active,
       -- i.e. is already inactive or expired\\:
       hyrex_scheduler_lock.is_active = false
       OR hyrex_scheduler_lock.release_at <= now()
   )
RETURNING lockid
"""


@dataclasses.dataclass()
class AcquireSchedulerLockParams:
    worker_name: str
    duration: datetime.timedelta


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def acquire_scheduler_lock(self, arg: AcquireSchedulerLockParams) -> Optional[int]:
        row = self._conn.execute(sqlalchemy.text(ACQUIRE_SCHEDULER_LOCK), {"p1": arg.worker_name, "p2": arg.duration}).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def acquire_scheduler_lock(self, arg: AcquireSchedulerLockParams) -> Optional[int]:
        row = (await self._conn.execute(sqlalchemy.text(ACQUIRE_SCHEDULER_LOCK), {"p1": arg.worker_name, "p2": arg.duration})).first()
        if row is None:
            return None
        return row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_task_def.sql
import dataclasses
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_TASK_DEF = """-- name: get_task_def \\:one
SELECT task_name, cron_expr, source_code, arg_schema, queue, priority, max_retries, timeout_seconds, last_updated FROM hyrex_task_def
WHERE task_name = :p1
"""


@dataclasses.dataclass()
class GetTaskDefParams:
    task_name: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_task_def(self, arg: GetTaskDefParams) -> Optional[models.HyrexTaskDef]:
        row = self._conn.execute(sqlalchemy.text(GET_TASK_DEF), {"p1": arg.task_name}).first()
        if row is None:
            return None
        return models.HyrexTaskDef(
            task_name=row[0],
            cron_expr=row[1],
            source_code=row[2],
            arg_schema=row[3],
            queue=row[4],
            priority=row[5],
            max_retries=row[6],
            timeout_seconds=row[7],
            last_updated=row[8],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_task_def(self, arg: GetTaskDefParams) -> Optional[models.HyrexTaskDef]:
        row = (await self._conn.execute(sqlalchemy.text(GET_TASK_DEF), {"p1": arg.task_name})).first()
        if row is None:
            return None
        return models.HyrexTaskDef(
            task_name=row[0],
            cron_expr=row[1],
            source_code=row[2],
            arg_schema=row[3],
            queue=row[4],
            priority=row[5],
            max_retries=row[6],
            timeout_seconds=row[7],
            last_updated=row[8],
        )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: update_hype_cron_job_run_status.sql
import dataclasses
from typing import Any

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


UPDATE_HYPE_CRON_JOB_RUN_STATUS = """-- name: update_hype_cron_job_run_status \\:exec
UPDATE hype_cron_job_run_details
SET status = :p1\\:\\:hype_cron_job_status_enum,
    start_time = CASE WHEN start_time IS NULL THEN NOW() ELSE start_time END,
    end_time = CASE WHEN :p1 IN ('SUCCESS', 'FAILED') THEN NOW() ELSE end_time END
WHERE runid = :p2
"""


@dataclasses.dataclass()
class UpdateHypeCronJobRunStatusParams:
    status: Any
    runid: int


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def update_hype_cron_job_run_status(self, arg: UpdateHypeCronJobRunStatusParams) -> None:
        self._conn.execute(sqlalchemy.text(UPDATE_HYPE_CRON_JOB_RUN_STATUS), {"p1": arg.status, "p2": arg.runid})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def update_hype_cron_job_run_status(self, arg: UpdateHypeCronJobRunStatusParams) -> None:
        await self._conn.execute(sqlalchemy.text(UPDATE_HYPE_CRON_JOB_RUN_STATUS), {"p1": arg.status, "p2": arg.runid})



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _06_create_hyrex_stats_task_status_counts_table_indexes.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_STATS_TASK_STATUS_COUNTS_TABLE_INDEXES = """-- name: create_stats_task_status_counts_table_indexes \\:exec
CREATE INDEX IF NOT EXISTS idx_hstsc_timepoint
    ON hyrex_stats_task_status_counts(timepoint)
"""


@dataclasses.dataclass()
class CreateStatsTaskStatusCountsTableIndexesParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_stats_task_status_counts_table_indexes(self, arg: CreateStatsTaskStatusCountsTableIndexesParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_STATS_TASK_STATUS_COUNTS_TABLE_INDEXES))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_stats_task_status_counts_table_indexes(self, arg: CreateStatsTaskStatusCountsTableIndexesParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_STATS_TASK_STATUS_COUNTS_TABLE_INDEXES))



import multiprocessing as mp
import tempfile
from datetime import datetime

# To ensure consistency between MacOS and Linux
try:
    mp.set_start_method("spawn")
except RuntimeError:
    # Context already set, which is fine
    pass


import cProfile
import importlib
import os
import pstats
import sys
from pathlib import Path

import typer

from hyrex import constants
from hyrex.env_vars import EnvVars
from hyrex.init_db import init_postgres_db
from hyrex.worker.logging import LogLevel
from hyrex.worker.root_process import WorkerRootProcess

cli = typer.Typer()


@cli.command()
def init_db(
    database_string: str = typer.Option(
        os.getenv(EnvVars.DATABASE_URL),
        "--database-string",
        help="Database connection string",
    )
):
    """
    Creates the tables for hyrex tasks/workers in the given Postgres database
    """
    if os.getenv(EnvVars.API_KEY) is not None:
        typer.echo(f"{EnvVars.API_KEY} is set. Skipping database initialization.")
        return

    if database_string:
        init_postgres_db(database_string)
        typer.echo("Hyrex tables initialized.")
        return

    typer.echo(
        f"Error: Database connection string must be provided either through the --database-string flag or the {EnvVars.DATABASE_URL} env variable."
    )


def validate_app_module_path(app_module_path):
    try:
        sys.path.append(str(Path.cwd()))
        module_path, instance_name = app_module_path.split(":")
        # Import the worker module
        app_module = importlib.import_module(module_path)
        app_instance = getattr(app_module, instance_name)
    except ModuleNotFoundError as e:
        typer.echo(f"Error: {e}")
        sys.exit(1)


# Profile a specific function
def profile_function(func, *args, **kwargs):
    profiler = cProfile.Profile()
    profiler.enable()
    result = func(*args, **kwargs)
    profiler.disable()

    # Create a temporary file with a timestamp
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    profile_dir = tempfile.gettempdir()  # Gets the system temp directory
    profile_path = os.path.join(profile_dir, f"hyrex_profile_{timestamp}.prof")

    # Save stats to the file
    stats = pstats.Stats(profiler)
    stats.dump_stats(profile_path)

    # Also print the top 20 time-consuming calls to console
    stats.sort_stats("cumulative")
    stats.print_stats(20)

    # Print the file path for reference
    print(f"\nProfile data saved to: {profile_path}")

    return result


@cli.command()
def run_worker(
    app_module_path: str = typer.Argument(..., help="Module path to the Hyrex app"),
    queue_pattern: str = typer.Option(
        constants.ANY_QUEUE,
        "--queue-pattern",
        "-q",
        help="Which queue(s) to pull tasks from. Glob patterns supported. Defaults to `*`",
    ),
    num_processes: int = typer.Option(
        8, "--num-processes", "-p", help="Number of executor processes to run"
    ),
    log_level: LogLevel = typer.Option(
        "INFO",
        "--log-level",
        "-l",
        help="Set the log level",
        case_sensitive=False,
        show_default=True,
        show_choices=True,
    ),
    profiling: bool = False,
):
    """
    Run a Hyrex worker for the specified app module path
    """

    database_url = os.environ.get(EnvVars.DATABASE_URL)

    if not database_url and not os.environ.get(EnvVars.API_KEY):
        raise EnvironmentError(
            f"Either {EnvVars.DATABASE_URL} (local) or {EnvVars.API_KEY} (Hyrex Cloud) must be set to run Hyrex worker."
        )

    # Try to initialize DB in worker
    if database_url:
        init_db(database_string=database_url)

    # Prevents HyrexRegistry instances from creating their own dispatchers
    os.environ[EnvVars.WORKER_PROCESS] = "true"

    validate_app_module_path(app_module_path)
    # TODO: Validate queue pattern?

    try:
        worker_root = WorkerRootProcess(
            log_level=log_level.upper(),
            app_module_path=app_module_path,
            queue_pattern=queue_pattern,
            num_processes=num_processes,
        )

        if profiling:
            profile_function(worker_root.run)  # Pass the function, not the result
        else:
            worker_root.run()

    except Exception as e:
        typer.echo(f"Error running worker: {e}")
        sys.exit(1)


@cli.command()
def studio(
    port: int = typer.Option(
        int(os.getenv("STUDIO_PORT", 1337)),
        "--port",
        "-p",
        help="Port to run the studio server on",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Enable verbose logging",
    ),
):
    """
    Run Hyrex Studio server for database inspection and query execution
    """
    database_url = os.environ.get(EnvVars.DATABASE_URL) or os.environ.get("PGURL")
    
    if not database_url:
        typer.echo(
            f"Error: Database connection string must be provided through HYREX_DATABASE_URL or PGURL environment variable."
        )
        sys.exit(1)
    
    # Set environment variables for the studio server
    os.environ["STUDIO_PORT"] = str(port)
    if verbose:
        os.environ["STUDIO_VERBOSE"] = "true"
    
    try:
        import uvicorn
    except ImportError:
        typer.echo("Error: uvicorn is required to run the studio server. Install it with: pip install uvicorn")
        sys.exit(1)
    
    try:
        from hyrex.hyrex_studio_server import app
        
        uvicorn.run(app, host="0.0.0.0", port=port, log_level="info" if verbose else "warning")
    except ImportError as e:
        typer.echo(f"Error importing studio server: {e}")
        sys.exit(1)
    except Exception as e:
        typer.echo(f"Error running studio server: {e}")
        sys.exit(1)


if __name__ == "__main__":
    cli()



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_cron_job_run_details.sql
import dataclasses
from typing import AsyncIterator, Iterator

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_CRON_JOB_RUN_DETAILS = """-- name: get_cron_job_run_details \\:many
SELECT 
    jobid,
    runid,
    command,
    status,
    schedule_time,
    start_time,
    end_time
FROM hyrex_cron_job_run_details
WHERE jobid = :p1
ORDER BY schedule_time DESC
LIMIT 100
"""


@dataclasses.dataclass()
class GetCronJobRunDetailsParams:
    jobid: int


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_cron_job_run_details(self, arg: GetCronJobRunDetailsParams) -> Iterator[models.HyrexCronJobRunDetail]:
        result = self._conn.execute(sqlalchemy.text(GET_CRON_JOB_RUN_DETAILS), {"p1": arg.jobid})
        for row in result:
            yield models.HyrexCronJobRunDetail(
                jobid=row[0],
                runid=row[1],
                command=row[2],
                status=row[3],
                schedule_time=row[4],
                start_time=row[5],
                end_time=row[6],
            )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_cron_job_run_details(self, arg: GetCronJobRunDetailsParams) -> AsyncIterator[models.HyrexCronJobRunDetail]:
        result = await self._conn.stream(sqlalchemy.text(GET_CRON_JOB_RUN_DETAILS), {"p1": arg.jobid})
        async for row in result:
            yield models.HyrexCronJobRunDetail(
                jobid=row[0],
                runid=row[1],
                command=row[2],
                status=row[3],
                schedule_time=row[4],
                start_time=row[5],
                end_time=row[6],
            )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _12_create_hyrex_task_run_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_TASK_RUN_TABLE = """-- name: create_task_run_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_task_run (
                                              id              UUID                        NOT NULL PRIMARY KEY,
                                              durable_id      UUID                        NOT NULL,
                                              root_id         UUID                        NOT NULL,
                                              parent_id       UUID,
                                              workflow_run_id UUID DEFAULT NULL,
                                              workflow_dependencies UUID[] DEFAULT NULL,
                                              task_name       VARCHAR                     NOT NULL,
                                              args            JSON                        NOT NULL,
                                              queue           VARCHAR                     NOT NULL,
                                              max_retries     SMALLINT                    NOT NULL,
                                              priority        SMALLINT                    NOT NULL,
                                              timeout_seconds INT                         DEFAULT NULL CHECK (timeout_seconds IS NULL OR timeout_seconds > 0),
                                              status          task_run_status             NOT NULL,
                                              attempt_number  SMALLINT                    NOT NULL,
                                              scheduled_start TIMESTAMP WITH TIME ZONE,
                                              executor_id     UUID,
                                              queued          TIMESTAMP WITH TIME ZONE,
                                              started         TIMESTAMP WITH TIME ZONE,
                                              finished        TIMESTAMP WITH TIME ZONE,
                                              last_heartbeat  TIMESTAMP WITH TIME ZONE,
                                              idempotency_key VARCHAR,
                                              log_link        VARCHAR
)
"""


@dataclasses.dataclass()
class CreateTaskRunTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_task_run_table(self, arg: CreateTaskRunTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_task_run_table(self, arg: CreateTaskRunTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_TASK_RUN_TABLE))



CREATE_WORKFLOW_TABLE = """
    CREATE TABLE IF NOT EXISTS hyrex_workflow
    (
        workflow_name  TEXT NOT NULL PRIMARY KEY,
        cron_expr      TEXT,
        source_code    TEXT,
        default_config JSON,
        arg_schema     JSON,
        dag_structure  JSON,
        last_updated   TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
    );
"""

UPSERT_WORKFLOW = """
INSERT INTO hyrex_workflow (workflow_name, cron_expr, source_code, dag_structure, arg_schema, default_config, last_updated)
VALUES ($1, $2, $3, $4, $5, $6, NOW())
ON CONFLICT (workflow_name)
DO UPDATE SET 
    cron_expr = EXCLUDED.cron_expr,
    source_code = EXCLUDED.source_code,
    dag_structure = EXCLUDED.dag_structure,
    arg_schema = EXCLUDED.arg_schema,
    default_config = EXCLUDED.default_config,
    last_updated = NOW();
"""

CREATE_WORKFLOW_RUN_TABLE = """
    DO $$
BEGIN
    IF NOT EXISTS (SELECT 1 
                   FROM pg_type 
                   WHERE typname = 'workflow_run_status' 
                     AND typnamespace = 'public'::regnamespace) THEN
        CREATE TYPE public.workflow_run_status AS ENUM (
            'success',
            'failed',
            'running',
            'up_for_cancel',
            'canceled',
            'asleep'
        );
    END IF;
END $$;

-- Create workflow run table
CREATE TABLE IF NOT EXISTS hyrex_workflow_run (
    id              UUID                        NOT NULL PRIMARY KEY,
    parent_id       UUID,
    workflow_name   VARCHAR                     NOT NULL,
    args            JSON                        NOT NULL,
    queue           VARCHAR                     NOT NULL,
    timeout_seconds INT                         DEFAULT NULL CHECK (timeout_seconds IS NULL OR timeout_seconds > 0),
    status          workflow_run_status         NOT NULL,
    scheduled_start TIMESTAMP WITH TIME ZONE,
    queued          TIMESTAMP WITH TIME ZONE,
    started         TIMESTAMP WITH TIME ZONE,
    finished        TIMESTAMP WITH TIME ZONE,
    last_heartbeat  TIMESTAMP WITH TIME ZONE,
    idempotency_key VARCHAR
);
"""

INSERT_WORKFLOW_RUN = """
INSERT INTO hyrex_workflow_run (
  id,
  parent_id,
  workflow_name,
  args,
  queue,
  timeout_seconds,
  status,
  queued,
  last_heartbeat,
  idempotency_key
)
VALUES ($1, NULL, $2, $3, $4, $5, 'running'::workflow_run_status, now(), now(), $6)
RETURNING id;
"""

# Returns (workflow_id, workflow_run_status)
SET_WORKFLOW_RUN_STATUS_BASED_ON_TASK_RUNS = """
    WITH latest_attempts AS (
        -- Get the latest attempt for each durable_id
        SELECT DISTINCT ON (durable_id)
            durable_id,
            workflow_run_id,
            status,
            attempt_number,
            max_retries
        FROM hyrex_task_run
        ORDER BY durable_id, attempt_number DESC
    ),
         workflow_statuses AS (
             SELECT w.id AS workflow_run_id,
                    CASE
                        -- If any task has failed on its final attempt, workflow is failed
                        WHEN EXISTS (
                            SELECT 1
                            FROM latest_attempts
                            WHERE latest_attempts.workflow_run_id = w.id
                              AND status = 'failed'
                              AND attempt_number >= max_retries
                        ) THEN 'failed'::workflow_run_status

                        -- If any task is still in progress, workflow is running
                        WHEN EXISTS (
                            SELECT 1
                            FROM latest_attempts
                            WHERE latest_attempts.workflow_run_id = w.id
                              AND status IN ('running', 'queued', 'waiting', 'up_for_cancel')
                        ) THEN 'running'::workflow_run_status

                        -- If all tasks are either success or skipped, workflow is success
                        WHEN NOT EXISTS (
                            SELECT 1
                            FROM latest_attempts
                            WHERE latest_attempts.workflow_run_id = w.id
                              AND status NOT IN ('success', 'skipped')
                        ) THEN 'success'::workflow_run_status

                        -- Handle other states (lost, canceled)
                        WHEN EXISTS (
                            SELECT 1
                            FROM latest_attempts
                            WHERE latest_attempts.workflow_run_id = w.id
                              AND status IN ('lost', 'canceled')
                        ) THEN 'failed'::workflow_run_status

                        -- Otherwise, keep current status
                        ELSE w.status
                        END AS new_status
             FROM hyrex_workflow_run w
             WHERE w.id = $1 -- Specific workflow
               AND w.status NOT IN ('success', 'failed') -- Only if not in terminal state
         )
    UPDATE hyrex_workflow_run w
    SET status   = ws.new_status,
        finished = CASE
                       WHEN ws.new_status IN ('success', 'failed')
                           THEN CURRENT_TIMESTAMP
                       ELSE w.finished
            END
    FROM workflow_statuses ws
    WHERE w.id = ws.workflow_run_id
    RETURNING w.id, w.status;
"""

ADVANCE_WORKFLOW_RUN = """
    WITH latest_attempts AS (
        -- Get the latest attempt for each durable_id
        SELECT DISTINCT ON (durable_id) durable_id,
                                        status
        FROM hyrex_task_run
        ORDER BY durable_id, attempt_number DESC),
         tasks_ready_to_queue AS (
             -- Find waiting tasks where all dependencies are successful
             SELECT t.id
             FROM hyrex_task_run t
             WHERE t.workflow_run_id = $1
               AND t.status = 'waiting'
               AND (
                 -- Handle both cases: no dependencies, or all dependencies are successful
                 t.workflow_dependencies IS NULL
                     OR NOT EXISTS (
                     -- Check if there are any dependencies that aren't successful
                     SELECT 1
                     FROM unnest(t.workflow_dependencies) AS dep_durable_id
                              LEFT JOIN latest_attempts la ON la.durable_id = dep_durable_id
                     WHERE la.status IS NULL
                        OR la.status != 'success')
                 ))
    UPDATE hyrex_task_run
    SET status = 'queued'
    WHERE id IN (SELECT id FROM tasks_ready_to_queue);
"""

GET_WORKFLOW_RUN_ARGS = """
    SELECT args
    FROM hyrex_workflow_run
    WHERE id = $1;
"""



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _01_create_executor_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_EXECUTOR_TABLE = """-- name: create_executor_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_executor
(
    id             UUID    NOT NULL PRIMARY KEY,
    name           VARCHAR NOT NULL,
    worker_name    VARCHAR NOT NULL,
    queue_pattern  VARCHAR NOT NULL,
    queues         VARCHAR[] NOT NULL,
    started        TIMESTAMP WITH TIME ZONE,
    stopped        TIMESTAMP WITH TIME ZONE,
    last_heartbeat TIMESTAMP WITH TIME ZONE,
    stats          JSON,
    status         executor_status NOT NULL DEFAULT 'UNKNOWN'
)
"""


@dataclasses.dataclass()
class CreateExecutorTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_executor_table(self, arg: CreateExecutorTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_EXECUTOR_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_executor_table(self, arg: CreateExecutorTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_EXECUTOR_TABLE))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: flush_all_data.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


FLUSH_ALL_DATA = """-- name: flush_all_data \\:exec
BEGIN
"""


@dataclasses.dataclass()
class FlushAllDataParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def flush_all_data(self, arg: FlushAllDataParams) -> None:
        self._conn.execute(sqlalchemy.text(FLUSH_ALL_DATA))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def flush_all_data(self, arg: FlushAllDataParams) -> None:
        await self._conn.execute(sqlalchemy.text(FLUSH_ALL_DATA))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_task_by_name.sql
import dataclasses
import datetime
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_TASK_BY_NAME = """-- name: get_task_by_name \\:one
SELECT 
    task_name,
    cron_expr,
    source_code,
    last_updated
FROM hyrex_task_def
WHERE task_name = :p1
"""


@dataclasses.dataclass()
class GetTaskByNameParams:
    task_name: str


@dataclasses.dataclass()
class GetTaskByNameRow:
    task_name: str
    cron_expr: Optional[str]
    source_code: Optional[str]
    last_updated: Optional[datetime.datetime]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_task_by_name(self, arg: GetTaskByNameParams) -> Optional[GetTaskByNameRow]:
        row = self._conn.execute(sqlalchemy.text(GET_TASK_BY_NAME), {"p1": arg.task_name}).first()
        if row is None:
            return None
        return GetTaskByNameRow(
            task_name=row[0],
            cron_expr=row[1],
            source_code=row[2],
            last_updated=row[3],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_task_by_name(self, arg: GetTaskByNameParams) -> Optional[GetTaskByNameRow]:
        row = (await self._conn.execute(sqlalchemy.text(GET_TASK_BY_NAME), {"p1": arg.task_name})).first()
        if row is None:
            return None
        return GetTaskByNameRow(
            task_name=row[0],
            cron_expr=row[1],
            source_code=row[2],
            last_updated=row[3],
        )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_app_name.sql
import dataclasses
from typing import Any, Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_APP_NAME = """-- name: get_app_name \\:one
SELECT app_info->>'name' as app_name
FROM hyrex_app
LIMIT 1
"""


@dataclasses.dataclass()
class GetAppNameParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_app_name(self, arg: GetAppNameParams) -> Optional[Optional[Any]]:
        row = self._conn.execute(sqlalchemy.text(GET_APP_NAME)).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_app_name(self, arg: GetAppNameParams) -> Optional[Optional[Any]]:
        row = (await self._conn.execute(sqlalchemy.text(GET_APP_NAME))).first()
        if row is None:
            return None
        return row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: kv_flush_keys.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


FLUSH_KEYS = """-- name: flush_keys \\:exec
DELETE FROM hyrex_kv
"""


@dataclasses.dataclass()
class FlushKeysParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def flush_keys(self, arg: FlushKeysParams) -> None:
        self._conn.execute(sqlalchemy.text(FLUSH_KEYS))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def flush_keys(self, arg: FlushKeysParams) -> None:
        await self._conn.execute(sqlalchemy.text(FLUSH_KEYS))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_workflow_runs_paginated.sql
import dataclasses
import datetime
from typing import Any, AsyncIterator, Iterator, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_WORKFLOW_RUNS_PAGINATED = """-- name: get_workflow_runs_paginated \\:many
SELECT 
    ROW_NUMBER() OVER (ORDER BY queued DESC) as row_number,
    id,
    parent_id,
    workflow_name,
    args,
    queue,
    timeout_seconds,
    status,
    scheduled_start,
    queued,
    started,
    finished,
    last_heartbeat,
    idempotency_key
FROM hyrex_workflow_run
ORDER BY queued DESC
LIMIT :p1 OFFSET :p2
"""


@dataclasses.dataclass()
class GetWorkflowRunsPaginatedParams:
    limit: int
    offset: int


@dataclasses.dataclass()
class GetWorkflowRunsPaginatedRow:
    row_number: int
    id: uuid.UUID
    parent_id: Optional[uuid.UUID]
    workflow_name: str
    args: Any
    queue: str
    timeout_seconds: Optional[int]
    status: Any
    scheduled_start: Optional[datetime.datetime]
    queued: Optional[datetime.datetime]
    started: Optional[datetime.datetime]
    finished: Optional[datetime.datetime]
    last_heartbeat: Optional[datetime.datetime]
    idempotency_key: Optional[str]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_workflow_runs_paginated(self, arg: GetWorkflowRunsPaginatedParams) -> Iterator[GetWorkflowRunsPaginatedRow]:
        result = self._conn.execute(sqlalchemy.text(GET_WORKFLOW_RUNS_PAGINATED), {"p1": arg.limit, "p2": arg.offset})
        for row in result:
            yield GetWorkflowRunsPaginatedRow(
                row_number=row[0],
                id=row[1],
                parent_id=row[2],
                workflow_name=row[3],
                args=row[4],
                queue=row[5],
                timeout_seconds=row[6],
                status=row[7],
                scheduled_start=row[8],
                queued=row[9],
                started=row[10],
                finished=row[11],
                last_heartbeat=row[12],
                idempotency_key=row[13],
            )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_workflow_runs_paginated(self, arg: GetWorkflowRunsPaginatedParams) -> AsyncIterator[GetWorkflowRunsPaginatedRow]:
        result = await self._conn.stream(sqlalchemy.text(GET_WORKFLOW_RUNS_PAGINATED), {"p1": arg.limit, "p2": arg.offset})
        async for row in result:
            yield GetWorkflowRunsPaginatedRow(
                row_number=row[0],
                id=row[1],
                parent_id=row[2],
                workflow_name=row[3],
                args=row[4],
                queue=row[5],
                timeout_seconds=row[6],
                status=row[7],
                scheduled_start=row[8],
                queued=row[9],
                started=row[10],
                finished=row[11],
                last_heartbeat=row[12],
                idempotency_key=row[13],
            )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: conditionally_retry_task.sql
import dataclasses
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CONDITIONALLY_RETRY_TASK = """-- name: conditionally_retry_task \\:one
SELECT result FROM conditionally_retry_task(
    :p1\\:\\:UUID,
    :p2\\:\\:UUID,
    :p3\\:\\:INT
) AS result
"""


@dataclasses.dataclass()
class ConditionallyRetryTaskParams:
    existing_task_id: uuid.UUID
    new_task_id: uuid.UUID
    timeout_seconds: Optional[int]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def conditionally_retry_task(self, arg: ConditionallyRetryTaskParams) -> Optional[Optional[Any]]:
        row = self._conn.execute(sqlalchemy.text(CONDITIONALLY_RETRY_TASK), {"p1": arg.existing_task_id, "p2": arg.new_task_id, "p3": arg.timeout_seconds}).first()
        if row is None:
            return None
        return row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def conditionally_retry_task(self, arg: ConditionallyRetryTaskParams) -> Optional[Optional[Any]]:
        row = (await self._conn.execute(sqlalchemy.text(CONDITIONALLY_RETRY_TASK), {"p1": arg.existing_task_id, "p2": arg.new_task_id, "p3": arg.timeout_seconds})).first()
        if row is None:
            return None
        return row[0]



import threading

_current_builder = threading.local()


def get_current_workflow_builder():
    return getattr(_current_builder, "builder", None)


def set_current_workflow_builder(builder):
    _current_builder.builder = builder


def clear_current_workflow_builder():
    _current_builder.builder = None



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_workflow_run_by_id.sql
import dataclasses
from typing import Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_WORKFLOW_RUN_BY_ID = """-- name: get_workflow_run_by_id \\:one
SELECT 
    id,
    parent_id,
    workflow_name,
    args,
    queue,
    timeout_seconds,
    status,
    scheduled_start,
    queued,
    started,
    finished,
    last_heartbeat,
    idempotency_key
FROM hyrex_workflow_run
WHERE id = :p1
"""


@dataclasses.dataclass()
class GetWorkflowRunByIdParams:
    id: uuid.UUID


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_workflow_run_by_id(self, arg: GetWorkflowRunByIdParams) -> Optional[models.HyrexWorkflowRun]:
        row = self._conn.execute(sqlalchemy.text(GET_WORKFLOW_RUN_BY_ID), {"p1": arg.id}).first()
        if row is None:
            return None
        return models.HyrexWorkflowRun(
            id=row[0],
            parent_id=row[1],
            workflow_name=row[2],
            args=row[3],
            queue=row[4],
            timeout_seconds=row[5],
            status=row[6],
            scheduled_start=row[7],
            queued=row[8],
            started=row[9],
            finished=row[10],
            last_heartbeat=row[11],
            idempotency_key=row[12],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_workflow_run_by_id(self, arg: GetWorkflowRunByIdParams) -> Optional[models.HyrexWorkflowRun]:
        row = (await self._conn.execute(sqlalchemy.text(GET_WORKFLOW_RUN_BY_ID), {"p1": arg.id})).first()
        if row is None:
            return None
        return models.HyrexWorkflowRun(
            id=row[0],
            parent_id=row[1],
            workflow_name=row[2],
            args=row[3],
            queue=row[4],
            timeout_seconds=row[5],
            status=row[6],
            scheduled_start=row[7],
            queued=row[8],
            started=row[9],
            finished=row[10],
            last_heartbeat=row[11],
            idempotency_key=row[12],
        )



import time

from pydantic import BaseModel


class DataPoint(BaseModel):
    value: float
    timestamp: int


class MinuteAverage(BaseModel):
    minute: int  # Unix timestamp rounded to minute
    average: float


class TimeSeriesAverager:
    def __init__(self):
        self.data_points: list[DataPoint] = []

    def _get_minute_timestamp(self, timestamp: int) -> int:
        # Round down to nearest minute
        return (timestamp // 60000) * 60000

    def _group_by_minute(self) -> dict[int, list[DataPoint]]:
        grouped_data: dict[int, list[DataPoint]] = {}

        for point in self.data_points:
            minute = self._get_minute_timestamp(point.timestamp)
            if minute not in grouped_data:
                grouped_data[minute] = []
            grouped_data[minute].append(point)

        return grouped_data

    def submit(self, x: float) -> None:
        current_time = int(time.time() * 1000)  # Convert to milliseconds
        self.data_points.append(DataPoint(value=x, timestamp=current_time))

    def get_time_series(self) -> list[MinuteAverage]:
        grouped_data = self._group_by_minute()
        result: list[MinuteAverage] = []

        # Sort minutes chronologically
        sorted_minutes = sorted(grouped_data.keys())

        for minute in sorted_minutes:
            points = grouped_data[minute]
            avg = sum(point.value for point in points) / len(points)

            result.append(MinuteAverage(minute=minute, average=avg))

        return result

    def get_average_for_minute(self, timestamp: int) -> MinuteAverage | None:
        minute = self._get_minute_timestamp(timestamp)
        grouped_data = self._group_by_minute()

        if minute not in grouped_data:
            return None

        points = grouped_data[minute]
        avg = sum(point.value for point in points) / len(points)

        return MinuteAverage(minute=minute, average=avg)

    def get_current_minute_average(self) -> MinuteAverage:
        current_minute = self._get_minute_timestamp(int(time.time() * 1000))
        result = self.get_average_for_minute(current_minute)
        if result is None:
            return MinuteAverage(minute=current_minute, average=0)
        return result

    def clear(self) -> None:
        self.data_points = []

    def prune_data_older_than(self, timestamp: int) -> None:
        self.data_points = [
            point for point in self.data_points if point.timestamp >= timestamp
        ]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_all_task_defs.sql
import dataclasses
from typing import AsyncIterator, Iterator

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_ALL_TASK_DEFS = """-- name: get_all_task_defs \\:many
SELECT task_name, cron_expr, source_code, arg_schema, queue, priority, max_retries, timeout_seconds, last_updated FROM hyrex_task_def
ORDER BY task_name
"""


@dataclasses.dataclass()
class GetAllTaskDefsParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_all_task_defs(self, arg: GetAllTaskDefsParams) -> Iterator[models.HyrexTaskDef]:
        result = self._conn.execute(sqlalchemy.text(GET_ALL_TASK_DEFS))
        for row in result:
            yield models.HyrexTaskDef(
                task_name=row[0],
                cron_expr=row[1],
                source_code=row[2],
                arg_schema=row[3],
                queue=row[4],
                priority=row[5],
                max_retries=row[6],
                timeout_seconds=row[7],
                last_updated=row[8],
            )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_all_task_defs(self, arg: GetAllTaskDefsParams) -> AsyncIterator[models.HyrexTaskDef]:
        result = await self._conn.stream(sqlalchemy.text(GET_ALL_TASK_DEFS))
        async for row in result:
            yield models.HyrexTaskDef(
                task_name=row[0],
                cron_expr=row[1],
                source_code=row[2],
                arg_schema=row[3],
                queue=row[4],
                priority=row[5],
                max_retries=row[6],
                timeout_seconds=row[7],
                last_updated=row[8],
            )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_hype_cron_job_by_name.sql
import dataclasses
from typing import Optional

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_HYPE_CRON_JOB_BY_NAME = """-- name: get_hype_cron_job_by_name \\:one
SELECT jobid, schedule, command_type, command_params, active, jobname, activated_at, scheduled_jobs_confirmed_until, should_backfill FROM hype_cron_job
WHERE jobname = :p1
"""


@dataclasses.dataclass()
class GetHypeCronJobByNameParams:
    jobname: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_hype_cron_job_by_name(self, arg: GetHypeCronJobByNameParams) -> Optional[models.HypeCronJob]:
        row = self._conn.execute(sqlalchemy.text(GET_HYPE_CRON_JOB_BY_NAME), {"p1": arg.jobname}).first()
        if row is None:
            return None
        return models.HypeCronJob(
            jobid=row[0],
            schedule=row[1],
            command_type=row[2],
            command_params=row[3],
            active=row[4],
            jobname=row[5],
            activated_at=row[6],
            scheduled_jobs_confirmed_until=row[7],
            should_backfill=row[8],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_hype_cron_job_by_name(self, arg: GetHypeCronJobByNameParams) -> Optional[models.HypeCronJob]:
        row = (await self._conn.execute(sqlalchemy.text(GET_HYPE_CRON_JOB_BY_NAME), {"p1": arg.jobname})).first()
        if row is None:
            return None
        return models.HypeCronJob(
            jobid=row[0],
            schedule=row[1],
            command_type=row[2],
            command_params=row[3],
            active=row[4],
            jobname=row[5],
            activated_at=row[6],
            scheduled_jobs_confirmed_until=row[7],
            should_backfill=row[8],
        )



import logging
import os
import signal
import socket
import threading
import time
from datetime import datetime, timezone
from multiprocessing import Process, Queue
from uuid import UUID

from uuid6 import uuid7

from hyrex import constants
from hyrex.env_vars import EnvVars
from hyrex.worker.admin import WorkerAdmin
from hyrex.worker.cron_scheduler import WorkerCronScheduler
from hyrex.worker.executor.executor import WorkerExecutor
from hyrex.worker.logging import LogLevel, init_logging
from hyrex.worker.messages.admin_messages import (
    ExecutorHeartbeatMessage,
    ExecutorStoppedMessage,
    NewExecutorMessage,
    TaskCanceledMessage,
    TaskHeartbeatMessage,
)
from hyrex.worker.messages.root_messages import (
    CancelTaskMessage,
    HeartbeatRequestMessage,
    SetExecutorTaskMessage,
    TaskRegistrationComplete,
)


def generate_worker_name():
    hostname = socket.gethostname()
    pid = os.getpid()
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    return f"hyrex-worker-{hostname}-{pid}-{timestamp}"


class WorkerRootProcess:
    def __init__(
        self,
        log_level: LogLevel,
        app_module_path: str,
        queue_pattern: str = None,
        num_processes: int = constants.DEFAULT_EXECUTOR_PROCESSES,
    ):
        self.log_level = log_level
        init_logging(log_level=log_level)
        self.logger = logging.getLogger(__name__)

        self.app_module_path = app_module_path
        self.queue_pattern = queue_pattern
        self.num_processes = num_processes
        self.worker_name = generate_worker_name()

        self.next_executor_number = 1
        # Has an executor registered all current tasks/workflows? Triggers cron scheduler launch
        self.task_registration_complete = False

        self._register_app = True

        self.heartbeat_requested = False

        self.running_on_platform: bool = os.environ.get(EnvVars.API_KEY) is not None

        self._stop_event = threading.Event()
        self.task_id_to_executor_id: dict[str, str] = {}
        self.executor_id_to_process: dict[str, Process] = {}
        self.admin_process: Process = None
        self.cron_scheduler_process: Process = None
        self.root_message_queue = Queue()
        self.admin_message_queue = Queue()

        self.setup_signal_handlers()

    def setup_signal_handlers(self):
        def signal_handler(signum, frame):
            signame = signal.Signals(signum).name
            self.logger.info(f"\nReceived {signame}. Starting graceful shutdown...")
            self._stop_event.set()

        # Register the handler for both SIGTERM and SIGINT
        signal.signal(signal.SIGTERM, signal_handler)
        signal.signal(signal.SIGINT, signal_handler)

    # Keep incrementing for each new executor
    def get_next_executor_name(self):
        name = "E" + str(self.next_executor_number)
        self.next_executor_number += 1
        return name

    def _spawn_executor(self):
        executor_id = uuid7()
        executor_process = WorkerExecutor(
            log_level=self.log_level,
            root_message_queue=self.root_message_queue,
            app_module_path=self.app_module_path,
            queue=self.queue_pattern,
            executor_id=executor_id,
            register_app=self._register_app,
            worker_name=self.worker_name,
            executor_name=self.get_next_executor_name(),
        )
        # Only register app/tasks once per worker.
        if self._register_app:
            self._register_app = False
        executor_process.start()
        self.executor_id_to_process[executor_id] = executor_process
        # Notify admin of new executor
        self.admin_message_queue.put(NewExecutorMessage(executor_id=executor_id))

    def check_executor_processes(self):
        # Check each executor and respawn if process has died.
        stopped_executors = []
        for executor_id, process in self.executor_id_to_process.items():
            if process.exitcode != None:
                self.logger.info(f"Process {process.pid} stopped. Cleaning up.")
                stopped_executors.append(executor_id)

        for executor_id in stopped_executors:
            # Now clean up everything
            self.executor_id_to_process[executor_id].join(timeout=1.0)
            del self.executor_id_to_process[executor_id]
            # Let message thread clear the task ID mapping to avoid concurrent access issues.
            self.root_message_queue.put(
                SetExecutorTaskMessage(executor_id=executor_id, task_id=None)
            )
            # Notify admin of stopped message
            self.admin_message_queue.put(
                ExecutorStoppedMessage(executor_id=executor_id)
            )

            # Replace the executor process
            self._spawn_executor()

    def _spawn_admin(self):
        admin = WorkerAdmin(
            root_message_queue=self.root_message_queue,
            admin_message_queue=self.admin_message_queue,
            log_level=self.log_level,
        )
        admin.start()
        self.admin_process = admin

    def check_admin_process(self):
        if not self.admin_process or not self.admin_process.is_alive():
            self.logger.warning("Admin process not running, respawning...")
            if self.admin_process:  # Clean up old process if it exists
                self.admin_process.join(timeout=1.0)
            self._spawn_admin()

    def _spawn_cron_scheduler(self):
        cron_scheduler = WorkerCronScheduler(
            log_level=self.log_level, worker_name=self.worker_name
        )
        cron_scheduler.start()
        self.cron_scheduler_process = cron_scheduler

    def check_cron_scheduler_process(self):
        if (
            not self.cron_scheduler_process
            or not self.cron_scheduler_process.is_alive()
        ):
            self.logger.warning("Cron scheduler process not running, respawning...")
            if self.cron_scheduler_process:
                self.cron_scheduler_process.join(timeout=1.0)
            self._spawn_cron_scheduler()

    def _message_listener(self):
        while True:
            # Blocking
            message = self.root_message_queue.get()

            if message == None:
                break

            if isinstance(message, CancelTaskMessage):
                self.cancel_running_task(message.task_id)
            elif isinstance(message, SetExecutorTaskMessage):
                self.set_executor_task(
                    executor_id=message.executor_id, task_id=message.task_id
                )
            elif isinstance(message, TaskRegistrationComplete):
                self.task_registration_complete = True
            elif isinstance(message, HeartbeatRequestMessage):
                self.heartbeat_requested = True

    def clear_executor_task(self, executor_id: UUID):
        for k, v in self.task_id_to_executor_id.items():
            if v == executor_id:
                del self.task_id_to_executor_id[k]
                break

    def set_executor_task(self, executor_id: UUID, task_id: UUID):
        self.clear_executor_task(executor_id=executor_id)
        # Add new mapping (unless task_id is None)
        if task_id:
            self.task_id_to_executor_id[task_id] = executor_id

    def cancel_running_task(self, task_id: UUID):
        executor_id = self.task_id_to_executor_id.get(task_id)
        if executor_id:
            executor_process = self.executor_id_to_process[executor_id]
            executor_process.kill()
            self.logger.info(f"Killed executor process to cancel task {task_id}")

            # Notify admin of successful termination
            self.admin_message_queue.put(
                ExecutorStoppedMessage(executor_id=executor_id)
            )
            self.admin_message_queue.put(TaskCanceledMessage(task_id=task_id))

    def send_heartbeats(self):
        self.logger.debug("Sending task and executor heartbeats.")

        self.admin_message_queue.put(
            ExecutorHeartbeatMessage(
                executor_ids=self.executor_id_to_process.keys(),
                timestamp=datetime.now(timezone.utc),
            )
        )
        self.admin_message_queue.put(
            TaskHeartbeatMessage(
                task_ids=self.task_id_to_executor_id.keys(),
                timestamp=datetime.now(timezone.utc),
            )
        )

    def run(self):
        self.message_listener_thread = threading.Thread(target=self._message_listener)
        self.message_listener_thread.start()
        self.logger.info("Incoming message queue now active...")

        self.logger.info("Spawning admin process.")
        self._spawn_admin()

        self.logger.info(f"Spawning {self.num_processes} task executor processes.")
        for _ in range(self.num_processes):
            self._spawn_executor()

        self.logger.info("Waiting for executor to complete task registration.")
        while not self.task_registration_complete and not self._stop_event.is_set():
            time.sleep(0.5)

        # Hyrex Cloud handles cron scheduling
        if not self.running_on_platform:
            self.logger.info("Spawning cron scheduler process.")
            self._spawn_cron_scheduler()

        last_heartbeat = time.monotonic()

        try:
            while not self._stop_event.is_set():
                # Check admin and restart if it has died
                self.check_admin_process()

                # Hyrex Cloud handles cron scheduling
                if not self.running_on_platform:
                    # Check cron scheduler and restart if it has died
                    self.check_cron_scheduler_process()

                # Check all executors and restart any that have died
                self.check_executor_processes()

                # Send heartbeat if requested or we're overdue
                current_time = time.monotonic()
                if (
                    self.heartbeat_requested
                    or current_time - last_heartbeat
                    > constants.WORKER_HEARTBEAT_FREQUENCY
                ):
                    self.send_heartbeats()
                    last_heartbeat = time.monotonic()
                    self.heartbeat_requested = False

                # Interruptible sleep
                self._stop_event.wait(1)

        finally:
            self.stop()

    def stop(self):
        try:
            # Stop all executors
            for executor_process in self.executor_id_to_process.values():
                executor_process._stop_event.set()
            for executor_process in self.executor_id_to_process.values():
                executor_process.join(timeout=constants.WORKER_EXECUTOR_PROCESS_TIMEOUT)
                if executor_process.is_alive():
                    self.logger.warning(
                        "Executor process did not exit cleanly, force killing."
                    )
                    executor_process.kill()
                    executor_process.join(timeout=1.0)
        except Exception as e:
            print(f"Error during executor shutdown: {e}")

        # Hyrex Cloud handles cron scheduling
        if not self.running_on_platform:
            try:
                self.logger.info("Stopping cron scheduler process.")
                self.cron_scheduler_process._stop_event.set()
                self.cron_scheduler_process.join(
                    timeout=constants.WORKER_CRON_SCHEDULER_PROCESS_TIMEOUT
                )
                if self.cron_scheduler_process.is_alive():
                    self.logger.warning(
                        "Cron scheduler process did not exit cleanly, force killing."
                    )
                    self.cron_scheduler_process.kill()
                    self.cron_scheduler_process.join(timeout=1.0)
            except Exception as e:
                print(f"Error during cron scheduler shutdown: {e}")

        try:
            # Stop admin
            self.logger.info("Stopping admin process.")
            self.admin_process._stop_event.set()
            self.admin_process.join(timeout=constants.WORKER_ADMIN_PROCESS_TIMEOUT)
            if self.admin_process.is_alive():
                self.logger.warning(
                    "Admin process did not exit cleanly, force killing."
                )
                self.admin_process.kill()
                self.admin_process.join(timeout=1.0)
        except Exception as e:
            print(f"Error during admin shutdown: {e}")

        try:
            # Stop internal message listener
            self.root_message_queue.put(None)
            self.message_listener_thread.join()
            if self.message_listener_thread.is_alive():
                self.logger.warning("Message listener thread did not exit cleanly.")
            else:
                self.logger.info("Message listener thread closed successfully.")

        except Exception as e:
            print(f"Error during main process shutdown: {e}")

        self.logger.info("Worker root process completed.")



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: fetch_active_queue_names.sql
import dataclasses
from typing import AsyncIterator, Iterator

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


FETCH_ACTIVE_QUEUE_NAMES = """-- name: fetch_active_queue_names \\:many
WITH distinct_queues AS (SELECT DISTINCT queue
                         FROM hyrex_task_run
                         WHERE status = 'QUEUED'
                           AND queue SIMILAR TO :p1),
     queue_count AS (SELECT COUNT(*) AS cnt
                     FROM distinct_queues)
SELECT queue
FROM (
         -- If count <= 100000, just select all queues
         SELECT dq.queue
         FROM distinct_queues dq,
              queue_count qc
         WHERE qc.cnt <= 100000

         UNION ALL

         -- If count > 100000, select a random subset
         SELECT queue
         FROM (SELECT dq.queue,
                      row_number() OVER (ORDER BY random()) AS rn
               FROM distinct_queues dq,
                    queue_count qc
               WHERE qc.cnt > 100000) sub
         WHERE rn <= 100000) final_result
"""


@dataclasses.dataclass()
class FetchActiveQueueNamesParams:
    queue_pattern: str


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def fetch_active_queue_names(self, arg: FetchActiveQueueNamesParams) -> Iterator[str]:
        result = self._conn.execute(sqlalchemy.text(FETCH_ACTIVE_QUEUE_NAMES), {"p1": arg.queue_pattern})
        for row in result:
            yield row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def fetch_active_queue_names(self, arg: FetchActiveQueueNamesParams) -> AsyncIterator[str]:
        result = await self._conn.stream(sqlalchemy.text(FETCH_ACTIVE_QUEUE_NAMES), {"p1": arg.queue_pattern})
        async for row in result:
            yield row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: fetch_task.sql
import dataclasses
import datetime
from typing import Any, List, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


FETCH_TASK = """-- name: fetch_task \\:one
WITH next_task AS (SELECT id
                   FROM hyrex_task_run
                   WHERE hyrex_task_run.queue = :p2
                     AND hyrex_task_run.status = 'QUEUED'\\:\\:task_run_status
                     AND hyrex_task_run.task_name = ANY(:p3\\:\\:VARCHAR[])
                   ORDER BY priority ASC, queued
                       FOR UPDATE SKIP LOCKED
                   LIMIT 1)
UPDATE hyrex_task_run AS ht
SET status      = 'RUNNING'\\:\\:task_run_status,
    started     = CURRENT_TIMESTAMP,
    executor_id = :p1
FROM next_task
WHERE ht.id = next_task.id
RETURNING ht.id
    , ht.durable_id
    , ht.root_id
    , ht.parent_id
    , ht.workflow_run_id
    , ht.task_name
    , ht.args
    , ht.queue
    , ht.attempt_number
    , ht.max_retries
    , ht.priority
    , ht.timeout_seconds
    , ht.scheduled_start
    , ht.queued
    , ht.started
"""


@dataclasses.dataclass()
class FetchTaskParams:
    executor_id: Optional[uuid.UUID]
    queue: str
    task_names: List[str]


@dataclasses.dataclass()
class FetchTaskRow:
    id: uuid.UUID
    durable_id: uuid.UUID
    root_id: uuid.UUID
    parent_id: Optional[uuid.UUID]
    workflow_run_id: Optional[uuid.UUID]
    task_name: str
    args: Any
    queue: str
    attempt_number: int
    max_retries: int
    priority: int
    timeout_seconds: Optional[int]
    scheduled_start: Optional[datetime.datetime]
    queued: Optional[datetime.datetime]
    started: Optional[datetime.datetime]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def fetch_task(self, arg: FetchTaskParams) -> Optional[FetchTaskRow]:
        row = self._conn.execute(sqlalchemy.text(FETCH_TASK), {"p1": arg.executor_id, "p2": arg.queue, "p3": arg.task_names}).first()
        if row is None:
            return None
        return FetchTaskRow(
            id=row[0],
            durable_id=row[1],
            root_id=row[2],
            parent_id=row[3],
            workflow_run_id=row[4],
            task_name=row[5],
            args=row[6],
            queue=row[7],
            attempt_number=row[8],
            max_retries=row[9],
            priority=row[10],
            timeout_seconds=row[11],
            scheduled_start=row[12],
            queued=row[13],
            started=row[14],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def fetch_task(self, arg: FetchTaskParams) -> Optional[FetchTaskRow]:
        row = (await self._conn.execute(sqlalchemy.text(FETCH_TASK), {"p1": arg.executor_id, "p2": arg.queue, "p3": arg.task_names})).first()
        if row is None:
            return None
        return FetchTaskRow(
            id=row[0],
            durable_id=row[1],
            root_id=row[2],
            parent_id=row[3],
            workflow_run_id=row[4],
            task_name=row[5],
            args=row[6],
            queue=row[7],
            attempt_number=row[8],
            max_retries=row[9],
            priority=row[10],
            timeout_seconds=row[11],
            scheduled_start=row[12],
            queued=row[13],
            started=row[14],
        )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: update_queues_on_executor.sql
import dataclasses
from typing import List
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


UPDATE_QUEUES_ON_EXECUTOR = """-- name: update_queues_on_executor \\:exec
UPDATE hyrex_executor
SET queues = :p1
WHERE id = :p2
"""


@dataclasses.dataclass()
class UpdateQueuesOnExecutorParams:
    queues: List[str]
    id: uuid.UUID


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def update_queues_on_executor(self, arg: UpdateQueuesOnExecutorParams) -> None:
        self._conn.execute(sqlalchemy.text(UPDATE_QUEUES_ON_EXECUTOR), {"p1": arg.queues, "p2": arg.id})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def update_queues_on_executor(self, arg: UpdateQueuesOnExecutorParams) -> None:
        await self._conn.execute(sqlalchemy.text(UPDATE_QUEUES_ON_EXECUTOR), {"p1": arg.queues, "p2": arg.id})



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_executor_status_enum.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_EXECUTOR_STATUS_ENUM = """-- name: create_executor_status_enum \\:exec
DO $$
BEGIN
    -- Create enum if it doesn't exist
    IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'executor_status') THEN
        CREATE TYPE executor_status AS ENUM ('SHUTDOWN', 'LOST', 'RUNNING', 'UNKNOWN');
    END IF;
END$$
"""


@dataclasses.dataclass()
class CreateExecutorStatusEnumParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_executor_status_enum(self, arg: CreateExecutorStatusEnumParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_EXECUTOR_STATUS_ENUM))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_executor_status_enum(self, arg: CreateExecutorStatusEnumParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_EXECUTOR_STATUS_ENUM))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: gateway.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'gateway.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import requests_pb2 as requests__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rgateway.proto\x12\x17hyrex.performanceserver\x1a\x1bgoogle/protobuf/empty.proto\x1a\x0erequests.proto2\xf6\x19\n\x0eGatewayService\x12@\n\x0eTestConnection\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\x12N\n\tWriteLogs\x12).hyrex.performanceserver.WriteLogsRequest\x1a\x16.google.protobuf.Empty\x12J\n\x07\x45nqueue\x12\'.hyrex.performanceserver.EnqueueRequest\x1a\x16.google.protobuf.Empty\x12\\\n\x07\x44\x65queue\x12\'.hyrex.performanceserver.DequeueRequest\x1a(.hyrex.performanceserver.DequeueResponse\x12\x62\n\tGetQueues\x12).hyrex.performanceserver.GetQueuesRequest\x1a*.hyrex.performanceserver.GetQueuesResponse\x12w\n\x10GetTaskRunStatus\x12\x30.hyrex.performanceserver.GetTaskRunStatusRequest\x1a\x31.hyrex.performanceserver.GetTaskRunStatusResponse\x12R\n\x0bMarkSuccess\x12+.hyrex.performanceserver.MarkSuccessRequest\x1a\x16.google.protobuf.Empty\x12P\n\nMarkFailed\x12*.hyrex.performanceserver.MarkFailedRequest\x1a\x16.google.protobuf.Empty\x12\x65\n\nGetTaskRun\x12*.hyrex.performanceserver.GetTaskRunRequest\x1a+.hyrex.performanceserver.GetTaskRunResponse\x12}\n\x12GetDurableTaskRuns\x12\x32.hyrex.performanceserver.GetDurableTaskRunsRequest\x1a\x33.hyrex.performanceserver.GetDurableTaskRunsResponse\x12T\n\x0cRetryTaskRun\x12,.hyrex.performanceserver.RetryTaskRunRequest\x1a\x16.google.protobuf.Empty\x12\\\n\x10TaskRunHeartbeat\x12\x30.hyrex.performanceserver.TaskRunHeartbeatRequest\x1a\x16.google.protobuf.Empty\x12\x66\n\x15TryToCancelDurableRun\x12\x35.hyrex.performanceserver.TryToCancelDurableRunRequest\x1a\x16.google.protobuf.Empty\x12i\n\x16GetTaskRunsUpForCancel\x12\x16.google.protobuf.Empty\x1a\x37.hyrex.performanceserver.GetTaskRunsUpForCancelResponse\x12T\n\x0cMarkCanceled\x12,.hyrex.performanceserver.MarkCanceledRequest\x1a\x16.google.protobuf.Empty\x12P\n\nKVStoreSet\x12*.hyrex.performanceserver.KVStoreSetRequest\x1a\x16.google.protobuf.Empty\x12\x65\n\nKVStoreGet\x12*.hyrex.performanceserver.KVStoreGetRequest\x1a+.hyrex.performanceserver.KVStoreGetResponse\x12V\n\rKVStoreDelete\x12-.hyrex.performanceserver.KVStoreDeleteRequest\x1a\x16.google.protobuf.Empty\x12\x64\n\x14MarkRunningTasksLost\x12\x34.hyrex.performanceserver.MarkRunningTasksLostRequest\x1a\x16.google.protobuf.Empty\x12Z\n\x0fRegisterTaskDef\x12/.hyrex.performanceserver.RegisterTaskDefRequest\x1a\x16.google.protobuf.Empty\x12\x65\n\nGetTaskDef\x12*.hyrex.performanceserver.GetTaskDefRequest\x1a+.hyrex.performanceserver.GetTaskDefResponse\x12q\n\x0eGetAllTaskDefs\x12..hyrex.performanceserver.GetAllTaskDefsRequest\x1a/.hyrex.performanceserver.GetAllTaskDefsResponse\x12\\\n\x10RegisterExecutor\x12\x30.hyrex.performanceserver.RegisterExecutorRequest\x1a\x16.google.protobuf.Empty\x12^\n\x11\x45xecutorHeartbeat\x12\x31.hyrex.performanceserver.ExecutorHeartbeatRequest\x1a\x16.google.protobuf.Empty\x12\x62\n\x13UpdateExecutorStats\x12\x33.hyrex.performanceserver.UpdateExecutorStatsRequest\x1a\x16.google.protobuf.Empty\x12\x64\n\x14UpdateExecutorQueues\x12\x34.hyrex.performanceserver.UpdateExecutorQueuesRequest\x1a\x16.google.protobuf.Empty\x12`\n\x12\x44isconnectExecutor\x12\x32.hyrex.performanceserver.DisconnectExecutorRequest\x1a\x16.google.protobuf.Empty\x12R\n\x0bRegisterApp\x12+.hyrex.performanceserver.RegisterAppRequest\x1a\x16.google.protobuf.Empty\x12P\n\nSetLogLink\x12*.hyrex.performanceserver.SetLogLinkRequest\x1a\x16.google.protobuf.Empty\x12\\\n\x10RegisterWorkflow\x12\x30.hyrex.performanceserver.RegisterWorkflowRequest\x1a\x16.google.protobuf.Empty\x12Z\n\x0fSendWorkflowRun\x12/.hyrex.performanceserver.SendWorkflowRunRequest\x1a\x16.google.protobuf.Empty\x12}\n\x12GetWorkflowRunArgs\x12\x32.hyrex.performanceserver.GetWorkflowRunArgsRequest\x1a\x33.hyrex.performanceserver.GetWorkflowRunArgsResponse\x12`\n\x12\x41\x64vanceWorkflowRun\x12\x32.hyrex.performanceserver.AdvanceWorkflowRunRequest\x1a\x16.google.protobuf.Empty\x12\x89\x01\n\x16GetWorkflowDurableRuns\x12\x36.hyrex.performanceserver.GetWorkflowDurableRunsRequest\x1a\x37.hyrex.performanceserver.GetWorkflowDurableRunsResponseB\x08Z\x06proto/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gateway_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\006proto/'
  _globals['_GATEWAYSERVICE']._serialized_start=88
  _globals['_GATEWAYSERVICE']._serialized_end=3406
# @@protoc_insertion_point(module_scope)



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: batch_update_heartbeat_log.sql
import dataclasses
from typing import List
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


BATCH_UPDATE_HEARTBEAT_LOG = """-- name: batch_update_heartbeat_log \\:exec
INSERT INTO hyrex_system_logs (id, timestamp, event_name, event_body)
VALUES (
    :p1\\:\\:uuid,
    NOW(),
    'BATCH_HEARTBEAT_UPDATE',
    json_build_object('updated_executor_ids', :p2\\:\\:uuid[])
)
"""


@dataclasses.dataclass()
class BatchUpdateHeartbeatLogParams:
    log_id: uuid.UUID
    executor_ids: List[uuid.UUID]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def batch_update_heartbeat_log(self, arg: BatchUpdateHeartbeatLogParams) -> None:
        self._conn.execute(sqlalchemy.text(BATCH_UPDATE_HEARTBEAT_LOG), {"p1": arg.log_id, "p2": arg.executor_ids})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def batch_update_heartbeat_log(self, arg: BatchUpdateHeartbeatLogParams) -> None:
        await self._conn.execute(sqlalchemy.text(BATCH_UPDATE_HEARTBEAT_LOG), {"p1": arg.log_id, "p2": arg.executor_ids})



# studio_server.py
import os
import sys
import asyncio
from pathlib import Path
from urllib.parse import urlparse
from contextlib import asynccontextmanager

from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from dotenv import load_dotenv
import asyncpg
from colorama import init as colorama_init, Fore, Style

# ------------------------------------------------------------
# 1. Environment & configuration
# ------------------------------------------------------------
BASE_DIR = Path(__file__).resolve().parent
load_dotenv(dotenv_path=BASE_DIR / ".env")  # same behaviour as dotenv.config()

PORT = int(os.getenv("STUDIO_PORT", 1337))
STUDIO_VERBOSE = os.getenv("STUDIO_VERBOSE", "false").lower() == "true"

DB_CONNECTION_STRING = os.getenv("HYREX_DATABASE_URL") or os.getenv(
    "PGURL"
)  # envVariables.getDatabaseUrl()‑equivalent
if not DB_CONNECTION_STRING:
    sys.stderr.write(
        "To run Hyrex Studio you must specify a database connection string "
        "(set HYREX_DATABASE_URL or PGURL).\n"
    )
    sys.exit(1)

# ------------------------------------------------------------
# 2. Colour / banner helpers
# ------------------------------------------------------------
colorama_init(autoreset=True)

COLOR_MAP = {
    "brightYellow": Fore.LIGHTYELLOW_EX,
    "brightWhite": Fore.LIGHTWHITE_EX,
    "brightMagenta": Fore.LIGHTMAGENTA_EX,
    "brightCyan": Fore.LIGHTCYAN_EX,
    "brightGreen": Fore.LIGHTGREEN_EX,
    "brightBlue": Fore.LIGHTBLUE_EX,
    "yellow": Fore.YELLOW,
    "dim": Style.DIM,
    "reset": Style.RESET_ALL,
}


def colorize(text: str, color: str) -> str:
    return COLOR_MAP.get(color, "") + text + COLOR_MAP["reset"]


def create_box(lines, box_color="brightCyan", width=55, style="double"):
    chars = {
        "double": dict(tl="╔", tr="╗", bl="╚", br="╝", h="═", v="║"),
        "single": dict(tl="┌", tr="┐", bl="└", br="┘", h="─", v="│"),
        "rounded": dict(tl="╭", tr="╮", bl="╰", br="╯", h="─", v="│"),
    }[style]

    top = chars["tl"] + chars["h"] * width + chars["tr"]
    bottom = chars["bl"] + chars["h"] * width + chars["br"]

    print(colorize("  " + top, box_color))
    for text, txt_color in lines:
        if text == "":
            line = chars["v"] + " " * width + chars["v"]
        else:
            pad = width - len(text)
            left = pad // 2
            right = pad - left
            line = (
                chars["v"]
                + " " * left
                + colorize(text, txt_color)
                + " " * right
                + chars["v"]
            )
        print(colorize("  " + line, box_color))
    print(colorize("  " + bottom, box_color))


def banner():
    db_name = urlparse(DB_CONNECTION_STRING).path.lstrip("/")
    os.system("cls" if os.name == "nt" else "clear")
    print()
    print(
        colorize("  ✨", "brightYellow"),
        colorize("Welcome to", "brightWhite"),
        colorize("✨", "brightYellow"),
    )
    create_box(
        [("", None), ("🚀 HYREX STUDIO SERVER 🚀", "brightMagenta"), ("", None)],
        "brightCyan",
        55,
        "double",
    )
    print()
    print(
        colorize("  ▸ ", "brightGreen")
        + colorize("Status:", "yellow")
        + " "
        + colorize("● Running", "brightGreen")
    )
    print(
        colorize("  ▸ ", "brightBlue")
        + colorize("Port:", "yellow")
        + " "
        + colorize(str(PORT), "brightWhite")
    )
    print(
        colorize("  ▸ ", "brightMagenta")
        + colorize("Database:", "yellow")
        + " "
        + colorize(db_name, "brightWhite")
    )
    print(
        colorize("  ▸ ", "brightCyan")
        + colorize("Verbose:", "yellow")
        + " "
        + colorize(
            "✓ Enabled" if STUDIO_VERBOSE else "✗ Disabled",
            "brightGreen" if STUDIO_VERBOSE else "dim",
        )
    )
    print()
    create_box(
        [
            ("", None),
            ("🌐 Open Hyrex Studio in your browser:", "brightWhite"),
            ("", None),
            ("👉 https://local.hyrex.studio 👈", "brightCyan"),
            ("", None),
        ],
        "brightBlue",
        55,
        "rounded",
    )
    if not STUDIO_VERBOSE:
        print(colorize("\n  Tip: Set STUDIO_VERBOSE=true to see detailed logs", "dim"))
    print("\n")


# ------------------------------------------------------------
# 3. Database pool – use asyncpg (async, high‑perf, server‑side prepared)
# ------------------------------------------------------------
_pool: asyncpg.pool.Pool | None = None


async def get_pool() -> asyncpg.pool.Pool:
    global _pool
    if _pool is None:
        _pool = await asyncpg.create_pool(
            dsn=DB_CONNECTION_STRING, min_size=1, max_size=10
        )
    return _pool


# ------------------------------------------------------------
# 4. FastAPI app & models
# ------------------------------------------------------------
class QueryPayload(BaseModel):
    query: str
    params: list | None = None


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Startup
    await get_pool()
    banner()
    yield
    # Shutdown
    global _pool
    if _pool:
        await _pool.close()
        if STUDIO_VERBOSE:
            print("Database pool closed")


app = FastAPI(
    title="Hyrex Studio",
    version="1.0",
    docs_url="/docs" if STUDIO_VERBOSE else None,
    redoc_url=None,
    lifespan=lifespan,
)

# CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # adjust if needed
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


# ------------------------------------------------------------
# 5. Endpoints
# ------------------------------------------------------------
@app.get("/health")
async def health():
    return {"status": "OK", "timestamp": str(asyncio.get_event_loop().time())}


@app.post("/api/query")
async def raw_query(payload: QueryPayload):
    if STUDIO_VERBOSE:
        print("Received query payload:", payload.model_dump_json(indent=2))

    if not payload.query:
        raise HTTPException(status_code=400, detail="Query is required")

    pool = await get_pool()
    try:
        async with pool.acquire() as conn:
            stmt = await conn.prepare(payload.query)
            rows = await stmt.fetch(*(payload.params or []))
            # Convert asyncpg Record objects → dict
            rows_dict = [dict(r) for r in rows]
            return {
                "rows": rows_dict,
                "rowCount": len(rows),
                "fields": [
                    {"name": a.name, "dataTypeID": a.type.oid}
                    for a in stmt.get_attributes()
                ],
            }
    except Exception as exc:
        if STUDIO_VERBOSE:
            print("Error executing query:", exc, file=sys.stderr)
        raise HTTPException(status_code=500, detail=str(exc))


# ------------------------------------------------------------
# 6. Entry‑point helper
# ------------------------------------------------------------
if __name__ == "__main__":
    # Run with:  python studio_server.py  (or better: `uvicorn studio_server:app --port 1337`)
    import uvicorn

    uvicorn.run("studio_server:app", host="0.0.0.0", port=PORT, reload=False)



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: update_workflow_run_status.sql
import dataclasses
from typing import Any
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


UPDATE_WORKFLOW_RUN_STATUS = """-- name: update_workflow_run_status \\:exec
UPDATE hyrex_workflow_run
SET status = :p1\\:\\:workflow_run_status,
    last_heartbeat = now(),
    finished = CASE 
        WHEN :p1\\:\\:workflow_run_status IN ('SUCCESS', 'FAILED', 'CANCELED') THEN now()
        ELSE finished
    END
WHERE id = :p2
"""


@dataclasses.dataclass()
class UpdateWorkflowRunStatusParams:
    status: Any
    id: uuid.UUID


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def update_workflow_run_status(self, arg: UpdateWorkflowRunStatusParams) -> None:
        self._conn.execute(sqlalchemy.text(UPDATE_WORKFLOW_RUN_STATUS), {"p1": arg.status, "p2": arg.id})


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def update_workflow_run_status(self, arg: UpdateWorkflowRunStatusParams) -> None:
        await self._conn.execute(sqlalchemy.text(UPDATE_WORKFLOW_RUN_STATUS), {"p1": arg.status, "p2": arg.id})



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: get_task_run_by_id.sql
import dataclasses
import datetime
from typing import Any, Optional
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


GET_TASK_RUN_BY_ID = """-- name: get_task_run_by_id \\:one
SELECT 
    tr.id,
    tr.durable_id,
    tr.root_id,
    tr.parent_id,
    tr.task_name,
    tr.args,
    tr.queue,
    tr.max_retries,
    tr.priority,
    tr.status,
    tr.attempt_number,
    tr.scheduled_start,
    tr.executor_id,
    tr.queued,
    tr.started,
    tr.finished,
    tr.log_link,
    t.source_code,
    res.result
FROM hyrex_task_run tr
LEFT JOIN hyrex_task_def t ON tr.task_name = t.task_name
LEFT JOIN hyrex_task_result res ON tr.id = res.task_id
WHERE tr.id = :p1
"""


@dataclasses.dataclass()
class GetTaskRunByIdParams:
    id: uuid.UUID


@dataclasses.dataclass()
class GetTaskRunByIdRow:
    id: uuid.UUID
    durable_id: uuid.UUID
    root_id: uuid.UUID
    parent_id: Optional[uuid.UUID]
    task_name: str
    args: Any
    queue: str
    max_retries: int
    priority: int
    status: Any
    attempt_number: int
    scheduled_start: Optional[datetime.datetime]
    executor_id: Optional[uuid.UUID]
    queued: Optional[datetime.datetime]
    started: Optional[datetime.datetime]
    finished: Optional[datetime.datetime]
    log_link: Optional[str]
    source_code: Optional[str]
    result: Optional[Any]


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def get_task_run_by_id(self, arg: GetTaskRunByIdParams) -> Optional[GetTaskRunByIdRow]:
        row = self._conn.execute(sqlalchemy.text(GET_TASK_RUN_BY_ID), {"p1": arg.id}).first()
        if row is None:
            return None
        return GetTaskRunByIdRow(
            id=row[0],
            durable_id=row[1],
            root_id=row[2],
            parent_id=row[3],
            task_name=row[4],
            args=row[5],
            queue=row[6],
            max_retries=row[7],
            priority=row[8],
            status=row[9],
            attempt_number=row[10],
            scheduled_start=row[11],
            executor_id=row[12],
            queued=row[13],
            started=row[14],
            finished=row[15],
            log_link=row[16],
            source_code=row[17],
            result=row[18],
        )


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def get_task_run_by_id(self, arg: GetTaskRunByIdParams) -> Optional[GetTaskRunByIdRow]:
        row = (await self._conn.execute(sqlalchemy.text(GET_TASK_RUN_BY_ID), {"p1": arg.id})).first()
        if row is None:
            return None
        return GetTaskRunByIdRow(
            id=row[0],
            durable_id=row[1],
            root_id=row[2],
            parent_id=row[3],
            task_name=row[4],
            args=row[5],
            queue=row[6],
            max_retries=row[7],
            priority=row[8],
            status=row[9],
            attempt_number=row[10],
            scheduled_start=row[11],
            executor_id=row[12],
            queued=row[13],
            started=row[14],
            finished=row[15],
            log_link=row[16],
            source_code=row[17],
            result=row[18],
        )



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: mark_lost_executors.sql
import dataclasses
import datetime
from typing import AsyncIterator, Iterator
import uuid

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


MARK_LOST_EXECUTORS = """-- name: mark_lost_executors \\:many
UPDATE hyrex_executor
SET status = 'LOST',
    stopped = last_heartbeat
WHERE status = 'RUNNING'
AND last_heartbeat < (CURRENT_TIMESTAMP - :p1\\:\\:INTERVAL)
AND stopped IS NULL
RETURNING id
"""


@dataclasses.dataclass()
class MarkLostExecutorsParams:
    timeout: datetime.timedelta


class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def mark_lost_executors(self, arg: MarkLostExecutorsParams) -> Iterator[uuid.UUID]:
        result = self._conn.execute(sqlalchemy.text(MARK_LOST_EXECUTORS), {"p1": arg.timeout})
        for row in result:
            yield row[0]


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def mark_lost_executors(self, arg: MarkLostExecutorsParams) -> AsyncIterator[uuid.UUID]:
        result = await self._conn.stream(sqlalchemy.text(MARK_LOST_EXECUTORS), {"p1": arg.timeout})
        async for row in result:
            yield row[0]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _08_create_workflow_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_WORKFLOW_TABLE = """-- name: create_workflow_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_workflow
(
    workflow_name TEXT NOT NULL PRIMARY KEY,
    cron_expr     TEXT,
    source_code   TEXT,
    dag_structure JSONB,
    last_updated  TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
)
"""


@dataclasses.dataclass()
class CreateWorkflowTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_workflow_table(self, arg: CreateWorkflowTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_WORKFLOW_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_workflow_table(self, arg: CreateWorkflowTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_WORKFLOW_TABLE))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: create_execute_queued_cron_job_func.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_EXECUTE_QUEUED_CRON_JOB_FUNCTION = """-- name: create_execute_queued_cron_job_function \\:exec
CREATE OR REPLACE FUNCTION execute_queued_cron_job()
RETURNS text AS
$$
DECLARE
    cmd              text;
    selected_runid   bigint;
    selected_jobid   bigint;
    start_ts         timestamp;
    job_name         text;
BEGIN
    -- 1) Grab the first queued command (row) and lock it, also fetching the jobid.
    SELECT command, runid, jobid
      INTO cmd, selected_runid, selected_jobid
      FROM hyrex_cron_job_run_details
     WHERE status = 'QUEUED'
       AND schedule_time <= now()
     ORDER BY schedule_time
     LIMIT 1
     FOR UPDATE SKIP LOCKED;

    IF FOUND THEN
        -- 2) Mark start_time right before execution.
        start_ts \\:= clock_timestamp();

        -- 3) Execute the command text.
        EXECUTE cmd;

        -- 4) Mark the end_time and set status.
        UPDATE hyrex_cron_job_run_details
           SET status      = 'SUCCESS',
               start_time  = start_ts,
               end_time    = clock_timestamp()
         WHERE runid = selected_runid
           AND status = 'QUEUED';

        -- 5) Retrieve the jobname using the selected jobid.
        SELECT jobname
          INTO job_name
          FROM hyrex_cron_job
         WHERE jobid = selected_jobid;

        RETURN 'executed ' || COALESCE(job_name, 'unknown job');
    END IF;

    RETURN 'not_found';

EXCEPTION
    WHEN OTHERS THEN
        -- In the event of an error, mark job as failed and record times.
        UPDATE hyrex_cron_job_run_details
           SET status      = 'FAILED',
               start_time  = COALESCE(start_ts, clock_timestamp()),
               end_time    = clock_timestamp()
         WHERE runid = selected_runid
           AND status = 'QUEUED';
        RAISE;
END;
$$
LANGUAGE plpgsql
"""


@dataclasses.dataclass()
class CreateExecuteQueuedCronJobFunctionParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_execute_queued_cron_job_function(self, arg: CreateExecuteQueuedCronJobFunctionParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_EXECUTE_QUEUED_CRON_JOB_FUNCTION))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_execute_queued_cron_job_function(self, arg: CreateExecuteQueuedCronJobFunctionParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_EXECUTE_QUEUED_CRON_JOB_FUNCTION))



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
# source: _10_create_workflow_run_table.sql
import dataclasses

import sqlalchemy
import sqlalchemy.ext.asyncio

from . import models


CREATE_WORKFLOW_RUN_TABLE = """-- name: create_workflow_run_table \\:exec
CREATE TABLE IF NOT EXISTS hyrex_workflow_run (
    id              UUID                        NOT NULL PRIMARY KEY,
    parent_id       UUID,
    workflow_name   VARCHAR                     NOT NULL,
    args            JSON                        NOT NULL,
    queue           VARCHAR                     NOT NULL,
    timeout_seconds INT                         DEFAULT NULL CHECK (timeout_seconds IS NULL OR timeout_seconds > 0),
    status          workflow_run_status             NOT NULL,
    scheduled_start TIMESTAMP WITH TIME ZONE,
    queued          TIMESTAMP WITH TIME ZONE,
    started         TIMESTAMP WITH TIME ZONE,
    finished        TIMESTAMP WITH TIME ZONE,
    last_heartbeat  TIMESTAMP WITH TIME ZONE,
    idempotency_key VARCHAR
)
"""


@dataclasses.dataclass()
class CreateWorkflowRunTableParams:

    pass

class Querier:
    def __init__(self, conn: sqlalchemy.engine.Connection):
        self._conn = conn

    def create_workflow_run_table(self, arg: CreateWorkflowRunTableParams) -> None:
        self._conn.execute(sqlalchemy.text(CREATE_WORKFLOW_RUN_TABLE))


class AsyncQuerier:
    def __init__(self, conn: sqlalchemy.ext.asyncio.AsyncConnection):
        self._conn = conn

    async def create_workflow_run_table(self, arg: CreateWorkflowRunTableParams) -> None:
        await self._conn.execute(sqlalchemy.text(CREATE_WORKFLOW_RUN_TABLE))



from datetime import datetime
from uuid import UUID

from pydantic import BaseModel


class NewExecutorMessage(BaseModel):
    executor_id: UUID


class ExecutorStoppedMessage(BaseModel):
    executor_id: UUID


class TaskCanceledMessage(BaseModel):
    task_id: UUID


class ExecutorHeartbeatMessage(BaseModel):
    timestamp: datetime
    executor_ids: list[UUID]


class TaskHeartbeatMessage(BaseModel):
    timestamp: datetime
    task_ids: list[UUID]



# Code generated by sqlc. DO NOT EDIT.
# versions:
#   sqlc v1.29.0
import dataclasses
import datetime
from typing import Any, List, Optional
import uuid


@dataclasses.dataclass()
class HypeCronJob:
    jobid: int
    schedule: Optional[str]
    command_type: Any
    command_params: Any
    active: bool
    jobname: str
    activated_at: Optional[datetime.datetime]
    scheduled_jobs_confirmed_until: Optional[datetime.datetime]
    should_backfill: Optional[bool]


@dataclasses.dataclass()
class HypeCronJobRunDetail:
    jobid: int
    runid: int
    command_type: Any
    command_params: Any
    status: Optional[Any]
    schedule_time: datetime.datetime
    start_time: Optional[datetime.datetime]
    end_time: Optional[datetime.datetime]


@dataclasses.dataclass()
class HyrexApp:
    id: int
    app_info: Optional[Any]


@dataclasses.dataclass()
class HyrexCronJob:
    jobid: int
    schedule: Optional[str]
    command: str
    active: bool
    jobname: str
    job_source: Any
    activated_at: Optional[datetime.datetime]
    scheduled_jobs_confirmed_until: Optional[datetime.datetime]
    should_backfill: Optional[bool]


@dataclasses.dataclass()
class HyrexCronJobRunDetail:
    jobid: int
    runid: int
    command: str
    status: Optional[Any]
    schedule_time: datetime.datetime
    start_time: Optional[datetime.datetime]
    end_time: Optional[datetime.datetime]


@dataclasses.dataclass()
class HyrexExecutor:
    id: uuid.UUID
    name: str
    worker_name: str
    queue_pattern: str
    queues: List[str]
    started: Optional[datetime.datetime]
    stopped: Optional[datetime.datetime]
    last_heartbeat: Optional[datetime.datetime]
    stats: Optional[Any]
    status: Any


@dataclasses.dataclass()
class HyrexKv:
    key: str
    value: str
    created_at: datetime.datetime


@dataclasses.dataclass()
class HyrexSchedulerLock:
    lockid: int
    worker_name: str
    acquired_at: datetime.datetime
    heartbeat_at: datetime.datetime
    release_at: datetime.datetime
    is_active: bool


@dataclasses.dataclass()
class HyrexStatsTaskStatusCount:
    timepoint: datetime.datetime
    queued: Optional[int]
    running: Optional[int]
    waiting: Optional[int]
    failed: Optional[int]
    success: Optional[int]
    lost: Optional[int]
    total: Optional[int]
    queued_delta: Optional[int]
    success_delta: Optional[int]
    failed_delta: Optional[int]
    lost_delta: Optional[int]


@dataclasses.dataclass()
class HyrexSystemLog:
    id: uuid.UUID
    timestamp: Optional[datetime.datetime]
    event_name: str
    event_body: Any


@dataclasses.dataclass()
class HyrexTaskDef:
    task_name: str
    cron_expr: Optional[str]
    source_code: Optional[str]
    arg_schema: Optional[Any]
    queue: Optional[str]
    priority: Optional[int]
    max_retries: Optional[int]
    timeout_seconds: Optional[int]
    last_updated: Optional[datetime.datetime]


@dataclasses.dataclass()
class HyrexTaskResult:
    task_id: uuid.UUID
    result: Optional[Any]
    created_at: Optional[datetime.datetime]


@dataclasses.dataclass()
class HyrexTaskRun:
    id: uuid.UUID
    durable_id: uuid.UUID
    root_id: uuid.UUID
    parent_id: Optional[uuid.UUID]
    workflow_run_id: Optional[uuid.UUID]
    workflow_dependencies: Optional[List[uuid.UUID]]
    task_name: str
    args: Any
    queue: str
    max_retries: int
    priority: int
    timeout_seconds: Optional[int]
    status: Any
    attempt_number: int
    scheduled_start: Optional[datetime.datetime]
    executor_id: Optional[uuid.UUID]
    queued: Optional[datetime.datetime]
    started: Optional[datetime.datetime]
    finished: Optional[datetime.datetime]
    last_heartbeat: Optional[datetime.datetime]
    idempotency_key: Optional[str]
    log_link: Optional[str]


@dataclasses.dataclass()
class HyrexWorkflow:
    workflow_name: str
    cron_expr: Optional[str]
    source_code: Optional[str]
    dag_structure: Optional[Any]
    last_updated: Optional[datetime.datetime]


@dataclasses.dataclass()
class HyrexWorkflowRun:
    id: uuid.UUID
    parent_id: Optional[uuid.UUID]
    workflow_name: str
    args: Any
    queue: str
    timeout_seconds: Optional[int]
    status: Any
    scheduled_start: Optional[datetime.datetime]
    queued: Optional[datetime.datetime]
    started: Optional[datetime.datetime]
    finished: Optional[datetime.datetime]
    last_heartbeat: Optional[datetime.datetime]
    idempotency_key: Optional[str]
